from sonusai.mixture.datatypes import ClassCount
from sonusai.mixture.datatypes import GeneralizedIDs
from sonusai.mixture.mixdb import MixtureDatabase


def get_class_count_from_mixids(mixdb: MixtureDatabase, mixids: GeneralizedIDs = "*") -> ClassCount:
    """Sums the class counts for given mixids"""
    total_class_count = [0] * mixdb.num_classes
    m_ids = mixdb.mixids_to_list(mixids)
    for m_id in m_ids:
        class_count = mixdb.mixture_class_count(m_id)
        for cl in range(mixdb.num_classes):
            total_class_count[cl] += class_count[cl]

    return total_class_count



from sonusai.mixture.datatypes import AugmentationRule
from sonusai.mixture.datatypes import AugmentedTarget
from sonusai.mixture.datatypes import TargetFile


def get_augmented_targets(
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    mixups: list[int] | None = None,
) -> list[AugmentedTarget]:
    from .augmentation import get_augmentation_indices_for_mixup
    from .augmentation import get_mixups

    if mixups is None:
        mixups = get_mixups(target_augmentations)

    augmented_targets: list[AugmentedTarget] = []
    for mixup in mixups:
        augmentation_indices = get_augmentation_indices_for_mixup(target_augmentations, mixup)
        for target_index in range(len(target_files)):
            for augmentation_index in augmentation_indices:
                augmented_targets.append(
                    AugmentedTarget(
                        target_id=target_index,
                        target_augmentation_id=augmentation_index,
                    )
                )

    return augmented_targets


def get_class_index_for_augmented_target(augmented_target: AugmentedTarget, targets: list[TargetFile]) -> list[int]:
    return targets[augmented_target.target_id].class_indices


def get_mixup_for_augmented_target(augmented_target: AugmentedTarget, augmentations: list[AugmentationRule]) -> int:
    return augmentations[augmented_target.target_augmentation_id].mixup


def get_target_ids_for_class_index(
    targets: list[TargetFile], class_index: int, allow_multiple: bool = False
) -> list[int]:
    """Get a list of target indices containing the given class index.

    If allow_multiple is True, then include targets that contain multiple class indices.
    """
    target_indices = set()
    for target_index, target in enumerate(targets):
        indices = target.class_indices
        if len(indices) == 1 or allow_multiple:
            for index in indices:
                if index == class_index + 1:
                    target_indices.add(target_index)

    return sorted(target_indices)


def get_augmented_target_ids_for_class_index(
    augmented_targets: list[AugmentedTarget],
    targets: list[TargetFile],
    augmentations: list[AugmentationRule],
    class_index: int,
    mixup: int,
    allow_multiple: bool = False,
) -> list[int]:
    """Get a list of augmented target indices containing the given class index.

    If allow_multiple is True, then include targets that contain multiple class indices.
    """
    augmented_target_ids = set()
    for augmented_target_id, augmented_target in enumerate(augmented_targets):
        if get_mixup_for_augmented_target(augmented_target=augmented_target, augmentations=augmentations) == mixup:
            indices = get_class_index_for_augmented_target(augmented_target=augmented_target, targets=targets)
            if len(indices) == 1 or allow_multiple:
                for index in indices:
                    if index == class_index + 1:
                        augmented_target_ids.add(augmented_target_id)

    return sorted(augmented_target_ids)


def get_augmented_target_ids_by_class(
    augmented_targets: list[AugmentedTarget],
    targets: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    mixup: int,
    num_classes: int,
) -> list[list[int]]:
    indices = []
    for idx in range(num_classes):
        indices.append(
            get_augmented_target_ids_for_class_index(
                augmented_targets=augmented_targets,
                targets=targets,
                augmentations=target_augmentations,
                class_index=idx,
                mixup=mixup,
            )
        )
    return indices


def get_target_augmentations_for_mixup(
    target_augmentations: list[AugmentationRule], mixup: int
) -> list[AugmentationRule]:
    """Get target augmentations for a given mixup value

    :param target_augmentations: List of target augmentation rules
    :param mixup: Mixup value of interest
    :return: Target augmentations
    """
    return [target_augmentation for target_augmentation in target_augmentations if target_augmentation.mixup == mixup]


def get_augmented_target_ids_for_mixup(
    augmented_targets: list[AugmentedTarget],
    targets: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    mixup: int,
    num_classes: int,
) -> list[list[int]]:
    from collections import deque
    from random import shuffle

    mixup_indices = []

    if mixup == 1:
        for index, augmented_target in enumerate(augmented_targets):
            if (
                get_mixup_for_augmented_target(
                    augmented_target=augmented_target,
                    augmentations=target_augmentations,
                )
                == 1
            ):
                mixup_indices.append([index])
        return mixup_indices

    augmented_target_ids_by_class = get_augmented_target_ids_by_class(
        augmented_targets=augmented_targets,
        targets=targets,
        target_augmentations=target_augmentations,
        mixup=mixup,
        num_classes=num_classes,
    )

    if mixup > num_classes:
        raise ValueError(f"Specified mixup, {mixup}, is greater than the number of classes, {num_classes}")

    de: deque[int] = deque()

    # Keep looping until not enough targets remain for mixup
    while sum([1 for x in augmented_target_ids_by_class if x]) >= mixup:
        # Need more class indices?
        if len(de) < mixup:
            # Only choose classes that still have data
            counts = [len(item) for item in augmented_target_ids_by_class]
            # Need to subtract out indices already in the deque
            for idx in de:
                counts[idx] -= 1
            indices = [idx for idx, val in enumerate(counts) if val > 0]
            shuffle(indices)
            # Keep shuffling if the deque is not empty and the first new index matches the last item
            # (so that a class does not appear twice in a mixup)
            while de and indices[0] == de[-1]:
                shuffle(indices)
            for index in indices:
                de.append(index)

        class_indices = [de.popleft() for _ in range(mixup)]

        target_indices = []
        for class_index in class_indices:
            target_indices.append(augmented_target_ids_by_class[class_index].pop())

        mixup_indices.append(target_indices)

    return mixup_indices



from collections import namedtuple

TruthConfigRecord = namedtuple("TruthConfigRecord", ["id", "name", "function", "stride_reduction", "config"])

TruthParametersRecord = namedtuple("TruthParametersRecord", ["id", "name", "parameters"])

TargetFileRecord = namedtuple("TargetFileRecord", ["id", "name", "samples", "class_indices", "level_type", "speaker_id"])

NoiseFileRecord = namedtuple("NoiseFileRecord", ["id", "name", "samples"])

TopRecord = namedtuple(
    "TopRecord",
    [
        "id",
        "version",
        "class_balancing",
        "feature",
        "noise_mix_mode",
        "num_classes",
        "seed",
        "mixid_width",
        "speaker_metadata_tiers",
        "textgrid_metadata_tiers",
    ],
)

ClassLabelRecord = namedtuple("ClassLabelRecord", ["id", "label"])

ClassWeightsThresholdRecord = namedtuple("ClassWeightsThresholdRecord", ["id", "threshold"])

ImpulseResponseFileRecord = namedtuple("ImpulseResponseFileRecord", ["id", "file"])

SpectralMaskRecord = namedtuple(
    "SpectralMaskRecord",
    ["id", "f_max_width", "f_num", "t_max_width", "t_num", "t_max_percent"],
)

TargetRecord = namedtuple("TargetRecord", ["id", "file_id", "augmentation"])

MixtureRecord = namedtuple(
    "MixtureRecord",
    [
        "id",
        "name",
        "noise_file_id",
        "noise_augmentation",
        "noise_offset",
        "noise_snr_gain",
        "random_snr",
        "snr",
        "samples",
        "spectral_mask_id",
        "spectral_mask_seed",
        "target_snr_gain",
    ],
)



from pyaaware import ForwardTransform
from pyaaware import InverseTransform

from sonusai.mixture.datatypes import AudioF
from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import Augmentation
from sonusai.mixture.datatypes import AugmentationRule
from sonusai.mixture.datatypes import EnergyT
from sonusai.mixture.datatypes import FeatureGeneratorConfig
from sonusai.mixture.datatypes import FeatureGeneratorInfo
from sonusai.mixture.datatypes import GeneralizedIDs
from sonusai.mixture.datatypes import Mixture
from sonusai.mixture.datatypes import NoiseFile
from sonusai.mixture.datatypes import SpeechMetadata
from sonusai.mixture.datatypes import Target
from sonusai.mixture.datatypes import TargetFile
from sonusai.mixture.datatypes import TransformConfig
from sonusai.mixture.db_datatypes import MixtureRecord
from sonusai.mixture.db_datatypes import TargetRecord
from sonusai.mixture.mixdb import MixtureDatabase


def generic_ids_to_list(num_ids: int, ids: GeneralizedIDs = "*") -> list[int]:
    """Resolve generalized IDs to a list of integers

    :param num_ids: Total number of indices
    :param ids: Generalized IDs
    :return: List of ID integers
    """
    all_ids = list(range(num_ids))

    if isinstance(ids, str):
        if ids == "*":
            return all_ids

        try:
            result = eval(f"{all_ids}[{ids}]")  # noqa: S307
            if isinstance(result, list):
                return result
            else:
                return [result]
        except NameError as e:
            raise ValueError(f"Empty ids {ids}: {e}") from e

    if isinstance(ids, range):
        result = list(ids)
    elif isinstance(ids, int):
        result = [ids]
    else:
        result = ids

    if not all(isinstance(x, int) and 0 <= x < num_ids for x in result):
        raise ValueError(f"Invalid entries in ids of {ids}")

    if not result:
        raise ValueError(f"Empty ids {ids}")

    return result


def get_feature_generator_info(
    fg_config: FeatureGeneratorConfig,
) -> FeatureGeneratorInfo:
    from dataclasses import asdict

    from pyaaware import FeatureGenerator

    from .datatypes import FeatureGeneratorInfo
    from .datatypes import TransformConfig

    fg = FeatureGenerator(**asdict(fg_config))

    return FeatureGeneratorInfo(
        decimation=fg.decimation,
        stride=fg.stride,
        step=fg.step,
        feature_parameters=fg.feature_parameters,
        ft_config=TransformConfig(
            length=fg.ftransform_length,
            overlap=fg.ftransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.ftransform_ttype,
        ),
        eft_config=TransformConfig(
            length=fg.eftransform_length,
            overlap=fg.eftransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.eftransform_ttype,
        ),
        it_config=TransformConfig(
            length=fg.itransform_length,
            overlap=fg.itransform_overlap,
            bin_start=fg.bin_start,
            bin_end=fg.bin_end,
            ttype=fg.itransform_ttype,
        ),
    )


def mixture_all_speech_metadata(mixdb: MixtureDatabase, mixture: Mixture) -> list[dict[str, SpeechMetadata]]:
    """Get a list of all speech metadata for the given mixture"""
    from praatio.utilities.constants import Interval

    from .datatypes import SpeechMetadata

    results: list[dict[str, SpeechMetadata]] = []
    for target in mixture.targets:
        data: dict[str, SpeechMetadata] = {}
        for tier in mixdb.speaker_metadata_tiers:
            data[tier] = mixdb.speaker(mixdb.target_file(target.file_id).speaker_id, tier)

        for tier in mixdb.textgrid_metadata_tiers:
            item = get_textgrid_tier_from_target_file(mixdb.target_file(target.file_id).name, tier)
            if isinstance(item, list):
                # Check for tempo augmentation and adjust Interval start and end data as needed
                entries = []
                for entry in item:
                    if target.augmentation.tempo is not None:
                        entries.append(
                            Interval(
                                entry.start / target.augmentation.tempo,
                                entry.end / target.augmentation.tempo,
                                entry.label,
                            )
                        )
                    else:
                        entries.append(entry)
                data[tier] = entries
            else:
                data[tier] = item
        results.append(data)

    return results


def mixture_metadata(mixdb: MixtureDatabase, m_id: int | None = None, mixture: Mixture | None = None) -> str:
    """Create a string of metadata for a Mixture

    :param mixdb: Mixture database
    :param m_id: Mixture ID
    :param mixture: Mixture record
    :return: String of metadata
    """
    if m_id is not None:
        mixture = mixdb.mixture(m_id)

    if mixture is None:
        raise ValueError("No mixture specified.")

    metadata = ""
    speech_metadata = mixture_all_speech_metadata(mixdb, mixture)
    for mi, target in enumerate(mixture.targets):
        target_file = mixdb.target_file(target.file_id)
        target_augmentation = target.augmentation
        metadata += f"target {mi} name: {target_file.name}\n"
        metadata += f"target {mi} augmentation: {target.augmentation.to_dict()}\n"
        metadata += f"target {mi} ir: {mixdb.impulse_response_file(target_augmentation.ir)}\n"
        metadata += f"target {mi} target_gain: {target.gain if not mixture.is_noise_only else 0}\n"
        metadata += f"target {mi} class indices: {target_file.class_indices}\n"
        for key in target_file.truth_configs:
            metadata += f"target {mi} truth '{key}' function: {target_file.truth_configs[key].function}\n"
            metadata += f"target {mi} truth '{key}' config:   {target_file.truth_configs[key].config}\n"
        for key in speech_metadata[mi]:
            metadata += f"target {mi} speech {key}: {speech_metadata[mi][key]}\n"
    noise = mixdb.noise_file(mixture.noise.file_id)
    noise_augmentation = mixture.noise.augmentation
    metadata += f"noise name: {noise.name}\n"
    metadata += f"noise augmentation: {noise_augmentation.to_dict()}\n"
    metadata += f"noise ir: {mixdb.impulse_response_file(noise_augmentation.ir)}\n"
    metadata += f"noise offset: {mixture.noise_offset}\n"
    metadata += f"snr: {mixture.snr}\n"
    metadata += f"random_snr: {mixture.snr.is_random}\n"
    metadata += f"samples: {mixture.samples}\n"
    metadata += f"target_snr_gain: {float(mixture.target_snr_gain)}\n"
    metadata += f"noise_snr_gain: {float(mixture.noise_snr_gain)}\n"

    return metadata


def write_mixture_metadata(mixdb: MixtureDatabase, m_id: int | None = None, mixture: Mixture | None = None) -> None:
    """Write mixture metadata to a text file

    :param mixdb: Mixture database
    :param m_id: Mixture ID
    :param mixture: Mixture record
    """
    from os.path import join

    if m_id is not None:
        name = mixdb.mixture(m_id).name
    elif mixture is not None:
        name = mixture.name
    else:
        raise ValueError("No mixture specified.")

    name = join(mixdb.location, "mixture", name, "metadata.txt")
    with open(file=name, mode="w") as f:
        f.write(mixture_metadata(mixdb, m_id, mixture))


def from_mixture(
    mixture: Mixture,
) -> tuple[str, int, str, int, float, bool, float, int, int, int, float]:
    return (
        mixture.name,
        mixture.noise.file_id,
        mixture.noise.augmentation.to_json(),
        mixture.noise_offset,
        mixture.noise_snr_gain,
        mixture.snr.is_random,
        mixture.snr,
        mixture.samples,
        mixture.spectral_mask_id,
        mixture.spectral_mask_seed,
        mixture.target_snr_gain,
    )


def to_mixture(entry: MixtureRecord, targets: list[Target]) -> Mixture:
    import json

    from sonusai.utils import dataclass_from_dict

    from .datatypes import Noise
    from .datatypes import UniversalSNR

    return Mixture(
        targets=targets,
        name=entry.name,
        noise=Noise(
            file_id=entry.noise_file_id,
            augmentation=dataclass_from_dict(Augmentation, json.loads(entry.noise_augmentation)),  # pyright: ignore [reportArgumentType]
        ),
        noise_offset=entry.noise_offset,
        noise_snr_gain=entry.noise_snr_gain,
        snr=UniversalSNR(is_random=entry.random_snr, value=entry.snr),
        samples=entry.samples,
        spectral_mask_id=entry.spectral_mask_id,
        spectral_mask_seed=entry.spectral_mask_seed,
        target_snr_gain=entry.target_snr_gain,
    )


def from_target(target: Target) -> tuple[int, str]:
    return target.file_id, target.augmentation.to_json()


def to_target(entry: TargetRecord) -> Target:
    import json

    from sonusai.utils import dataclass_from_dict

    from .datatypes import Augmentation

    return Target(
        file_id=entry.file_id,
        augmentation=dataclass_from_dict(Augmentation, json.loads(entry.augmentation)),  # pyright: ignore [reportArgumentType]
    )


def get_target(mixdb: MixtureDatabase, mixture: Mixture, targets_audio: list[AudioT], use_cache: bool = True) -> AudioT:
    """Get the augmented target audio data for the given mixture record

    :param mixdb: Mixture database
    :param mixture: Mixture record
    :param targets_audio: List of augmented target audio data (one per target in the mixup)
    :param use_cache: If true, use LRU caching
    :return: Sum of augmented target audio data
    """
    # Apply impulse responses to targets
    import numpy as np

    from .audio import read_ir
    from .augmentation import apply_impulse_response

    targets_ir = []
    for idx, target in enumerate(targets_audio):
        ir_idx = mixture.targets[idx].augmentation.ir
        if ir_idx is not None:
            targets_ir.append(
                apply_impulse_response(audio=target, ir=read_ir(mixdb.impulse_response_file(int(ir_idx)), use_cache))  # pyright: ignore [reportArgumentType]
            )
        else:
            targets_ir.append(target)

    # Return sum of targets
    return np.sum(targets_ir, axis=0)


def get_transform_from_audio(audio: AudioT, transform: ForwardTransform) -> tuple[AudioF, EnergyT]:
    """Apply forward transform to input audio data to generate transform data

    :param audio: Time domain data [samples]
    :param transform: ForwardTransform object
    :return: Frequency domain data [frames, bins], Energy [frames]
    """
    import torch

    f, e = transform.execute_all(torch.from_numpy(audio))

    return f.numpy(), e.numpy()


def forward_transform(audio: AudioT, config: TransformConfig) -> AudioF:
    """Transform time domain data into frequency domain using the forward transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to forward_transform().

    :param audio: Time domain data [samples]
    :param config: Transform configuration
    :return: Frequency domain data [frames, bins]
    """
    from pyaaware import ForwardTransform

    audio_f, _ = get_transform_from_audio(
        audio=audio,
        transform=ForwardTransform(
            length=config.length,
            overlap=config.overlap,
            bin_start=config.bin_start,
            bin_end=config.bin_end,
            ttype=config.ttype,
        ),
    )
    return audio_f


def get_audio_from_transform(data: AudioF, transform: InverseTransform) -> tuple[AudioT, EnergyT]:
    """Apply inverse transform to input transform data to generate audio data

    :param data: Frequency domain data [frames, bins]
    :param transform: InverseTransform object
    :return: Time domain data [samples], Energy [frames]
    """

    import torch

    t, e = transform.execute_all(torch.from_numpy(data))

    return t.numpy(), e.numpy()


def inverse_transform(transform: AudioF, config: TransformConfig) -> AudioT:
    """Transform frequency domain data into time domain using the inverse transform config from the feature

    A new transform is used for each call; i.e., state is not maintained between calls to inverse_transform().

    :param transform: Frequency domain data [frames, bins]
    :param config: Transform configuration
    :return: Time domain data [samples]
    """
    from pyaaware import InverseTransform

    audio, _ = get_audio_from_transform(
        data=transform,
        transform=InverseTransform(
            length=config.length,
            overlap=config.overlap,
            bin_start=config.bin_start,
            bin_end=config.bin_end,
            ttype=config.ttype,
            gain=1,
        ),
    )
    return audio


def check_audio_files_exist(mixdb: MixtureDatabase) -> None:
    """Walk through all the noise and target audio files in a mixture database ensuring that they exist"""
    from os.path import exists

    from .tokenized_shell_vars import tokenized_expand

    for noise in mixdb.noise_files:
        file_name, _ = tokenized_expand(noise.name)
        if not exists(file_name):
            raise OSError(f"Could not find {file_name}")

    for target in mixdb.target_files:
        file_name, _ = tokenized_expand(target.name)
        if not exists(file_name):
            raise OSError(f"Could not find {file_name}")


def augmented_target_samples(
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    feature_step_samples: int,
) -> int:
    from itertools import product

    from .augmentation import estimate_augmented_length_from_length

    target_ids = list(range(len(target_files)))
    target_augmentation_ids = list(range(len(target_augmentations)))
    it = list(product(*[target_ids, target_augmentation_ids]))
    return sum(
        [
            estimate_augmented_length_from_length(
                length=target_files[fi].samples,
                tempo=target_augmentations[ai].tempo,
                frame_length=feature_step_samples,
            )
            for fi, ai in it
        ]
    )


def augmented_noise_samples(noise_files: list[NoiseFile], noise_augmentations: list[Augmentation]) -> int:
    from itertools import product

    noise_ids = list(range(len(noise_files)))
    noise_augmentation_ids = list(range(len(noise_augmentations)))
    it = list(product(*[noise_ids, noise_augmentation_ids]))
    return sum([augmented_noise_length(noise_files[fi], noise_augmentations[ai]) for fi, ai in it])


def augmented_noise_length(noise_file: NoiseFile, noise_augmentation: Augmentation) -> int:
    from .augmentation import estimate_augmented_length_from_length

    return estimate_augmented_length_from_length(length=noise_file.samples, tempo=noise_augmentation.tempo)


def get_textgrid_tier_from_target_file(target_file: str, tier: str) -> SpeechMetadata | None:
    from pathlib import Path

    from praatio import textgrid
    from praatio.utilities.constants import Interval

    from .tokenized_shell_vars import tokenized_expand

    textgrid_file = Path(tokenized_expand(target_file)[0]).with_suffix(".TextGrid")
    if not textgrid_file.exists():
        return None

    tg = textgrid.openTextgrid(str(textgrid_file), includeEmptyIntervals=False)

    if tier not in tg.tierNames:
        return None

    entries = tg.getTier(tier).entries
    if len(entries) > 1:
        return [entry for entry in entries if isinstance(entry, Interval)]

    if len(entries) == 1:
        return entries[0].label

    return None


def frames_from_samples(samples: int, step_samples: int) -> int:
    import numpy as np

    return int(np.ceil(samples / step_samples))



"""sonusai genmix

usage: genmix [-hvgtsn] [-i MIXID] LOC

options:
    -h, --help
    -v, --verbose               Be verbose.
    -i MIXID, --mixid MIXID     Mixture ID(s) to generate. [default: *].
    -g, --target                Save target. [default: False].
    -t, --truth                 Save truth_t. [default: False].
    -s, --segsnr                Save segsnr_t. [default: False].
    -n, --nopar                 Do not run in parallel. [default: False].

Generate SonusAI mixture data from a SonusAI mixture database.

Inputs:
    LOC         A SonusAI mixture database directory.
    MIXID       A glob of mixture ID(s) to generate.

Outputs the following to the mixture database directory:
    <id>
        mixture.pkl
        targets.pkl
        noise.pkl
        target.pkl (optional)
        truth_t.pkl (optional)
        segsnr_t.pkl (optional)
        metadata.txt
    genmix.log
"""

import signal

from sonusai.mixture import GeneralizedIDs
from sonusai.mixture import GenMixData


def signal_handler(_sig, _frame):
    import sys

    from sonusai import logger

    logger.info("Canceled due to keyboard interrupt")
    sys.exit(1)


signal.signal(signal.SIGINT, signal_handler)


def genmix(
    location: str,
    mixids: GeneralizedIDs = "*",
    save_target: bool = False,
    compute_truth: bool = False,
    compute_segsnr: bool = False,
    write: bool = False,
    show_progress: bool = False,
    force: bool = True,
    no_par: bool = False,
) -> list[GenMixData]:
    from functools import partial

    from sonusai.mixture import MixtureDatabase
    from sonusai.utils import par_track
    from sonusai.utils import track

    mixdb = MixtureDatabase(location)
    mixids = mixdb.mixids_to_list(mixids)
    progress = track(total=len(mixids), disable=not show_progress)
    results = par_track(
        partial(
            _genmix_kernel,
            location=location,
            save_target=save_target,
            compute_truth=compute_truth,
            compute_segsnr=compute_segsnr,
            force=force,
            write=write,
        ),
        mixids,
        progress=progress,
        no_par=no_par,
    )
    progress.close()

    return results


def _genmix_kernel(
    m_id: int,
    location: str,
    save_target: bool,
    compute_truth: bool,
    compute_segsnr: bool,
    force: bool,
    write: bool,
) -> GenMixData:
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import write_cached_data
    from sonusai.mixture import write_mixture_metadata

    mixdb = MixtureDatabase(location)

    result = GenMixData()

    targets = mixdb.mixture_targets(m_id=m_id, force=force)
    result.targets = targets
    noise = mixdb.mixture_noise(m_id=m_id, force=force)
    result.noise = noise
    if write:
        write_cached_data(
            mixdb.location,
            "mixture",
            mixdb.mixture(m_id).name,
            [
                ("targets", targets),
                ("noise", noise),
            ],
        )

    if compute_truth:
        truth_t = mixdb.mixture_truth_t(m_id=m_id, targets=targets, noise=noise, force=force)
        result.truth_t = truth_t
        if write:
            write_cached_data(mixdb.location, "mixture", mixdb.mixture(m_id).name, [("truth_t", truth_t)])

    target = mixdb.mixture_target(m_id=m_id, targets=targets)
    result.target = target
    if save_target and write:
        write_cached_data(mixdb.location, "mixture", mixdb.mixture(m_id).name, [("target", target)])

    if compute_segsnr:
        segsnr_t = mixdb.mixture_segsnr_t(m_id=m_id, targets=targets, target=target, noise=noise, force=force)
        result.segsnr_t = segsnr_t
        if write:
            write_cached_data(mixdb.location, "mixture", mixdb.mixture(m_id).name, [("segsnr_t", segsnr_t)])

    mixture = mixdb.mixture_mixture(m_id=m_id, targets=targets, target=target, noise=noise, force=force)
    result.mixture = mixture
    if write:
        write_cached_data(mixdb.location, "mixture", mixdb.mixture(m_id).name, [("mixture", mixture)])
        write_mixture_metadata(mixdb, m_id=m_id)

    return result


def main() -> None:
    from docopt import docopt

    import sonusai
    from sonusai.utils import trim_docstring

    args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

    import time
    from os.path import join

    from sonusai import create_file_handler
    from sonusai import initial_log_messages
    from sonusai import logger
    from sonusai import update_console_handler
    from sonusai.mixture import SAMPLE_RATE
    from sonusai.mixture import MixtureDatabase
    from sonusai.mixture import check_audio_files_exist
    from sonusai.utils import human_readable_size
    from sonusai.utils import seconds_to_hms

    verbose = args["--verbose"]
    location = args["LOC"]
    mixids = args["--mixid"]
    save_target = args["--target"]
    compute_truth = args["--truth"]
    compute_segsnr = args["--segsnr"]
    no_par = args["--nopar"]

    start_time = time.monotonic()

    create_file_handler(join(location, "genmix.log"))
    update_console_handler(verbose)
    initial_log_messages("genmix")

    logger.info(f"Load mixture database from {location}")
    mixdb = MixtureDatabase(location)
    mixids = mixdb.mixids_to_list(mixids)

    total_samples = mixdb.total_samples(mixids)
    duration = total_samples / SAMPLE_RATE

    logger.info("")
    logger.info(f"Found {len(mixids):,} mixtures to process")
    logger.info(f"{total_samples:,} samples")

    check_audio_files_exist(mixdb)

    try:
        genmix(
            location=location,
            mixids=mixids,
            save_target=save_target,
            compute_truth=compute_truth,
            compute_segsnr=compute_segsnr,
            write=True,
            show_progress=True,
            no_par=no_par,
        )
    except Exception as e:
        logger.debug(e)
        raise

    logger.info(f"Wrote {len(mixids)} mixtures to {location}")
    logger.info("")
    logger.info(f"Duration: {seconds_to_hms(seconds=duration)}")
    logger.info(f"mixture:  {human_readable_size(total_samples * 2, 1)}")
    if compute_truth:
        logger.info(f"truth_t:  {human_readable_size(total_samples * mixdb.num_classes * 4, 1)}")
    logger.info(f"target:   {human_readable_size(total_samples * 2, 1)}")
    logger.info(f"noise:    {human_readable_size(total_samples * 2, 1)}")
    if compute_segsnr:
        logger.info(f"segsnr:   {human_readable_size(total_samples * 4, 1)}")

    end_time = time.monotonic()
    logger.info(f"Completed in {seconds_to_hms(seconds=end_time - start_time)}")
    logger.info("")


if __name__ == "__main__":
    main()



from sonusai.mixture.datatypes import AugmentationRule
from sonusai.mixture.datatypes import AugmentedTarget
from sonusai.mixture.datatypes import TargetFile


def balance_targets(
    augmented_targets: list[AugmentedTarget],
    targets: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    class_balancing_augmentation: AugmentationRule,
    num_classes: int,
    num_ir: int,
    mixups: list[int] | None = None,
) -> tuple[list[AugmentedTarget], list[AugmentationRule]]:
    import math

    from .augmentation import get_mixups
    from .datatypes import AugmentedTarget
    from .targets import get_augmented_target_ids_by_class

    first_cba_id = len(target_augmentations)

    if mixups is None:
        mixups = get_mixups(target_augmentations)

    for mixup in mixups:
        if mixup == 1:
            continue

        augmented_target_indices_by_class = get_augmented_target_ids_by_class(
            augmented_targets=augmented_targets,
            targets=targets,
            target_augmentations=target_augmentations,
            mixup=mixup,
            num_classes=num_classes,
        )

        largest = max([len(item) for item in augmented_target_indices_by_class])
        largest = math.ceil(largest / mixup) * mixup
        for at_indices in augmented_target_indices_by_class:
            additional_augmentations_needed = largest - len(at_indices)
            target_ids = sorted({augmented_targets[at_index].target_id for at_index in at_indices})

            tfi_idx = 0
            for _ in range(additional_augmentations_needed):
                target_id = target_ids[tfi_idx]
                tfi_idx = (tfi_idx + 1) % len(target_ids)
                augmentation_index, target_augmentations = _get_unused_balancing_augmentation(
                    augmented_targets=augmented_targets,
                    targets=targets,
                    target_augmentations=target_augmentations,
                    class_balancing_augmentation=class_balancing_augmentation,
                    target_id=target_id,
                    mixup=mixup,
                    num_ir=num_ir,
                    first_cba_id=first_cba_id,
                )
                augmented_target = AugmentedTarget(target_id=target_id, target_augmentation_id=augmentation_index)
                augmented_targets.append(augmented_target)

    return augmented_targets, target_augmentations


def _get_unused_balancing_augmentation(
    augmented_targets: list[AugmentedTarget],
    targets: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    class_balancing_augmentation: AugmentationRule,
    target_id: int,
    mixup: int,
    num_ir: int,
    first_cba_id: int,
) -> tuple[int, list[AugmentationRule]]:
    """Get an unused balancing augmentation for a given target file index"""
    from dataclasses import asdict

    from .augmentation import get_augmentation_rules

    balancing_augmentations = [item for item in range(len(target_augmentations)) if item >= first_cba_id]
    used_balancing_augmentations = [
        at.target_augmentation_id
        for at in augmented_targets
        if at.target_id == target_id and at.target_augmentation_id in balancing_augmentations
    ]

    augmentation_indices = [
        item
        for item in balancing_augmentations
        if item not in used_balancing_augmentations and target_augmentations[item].mixup == mixup
    ]
    if len(augmentation_indices) > 0:
        return augmentation_indices[0], target_augmentations

    class_balancing_augmentation = get_class_balancing_augmentation(
        target=targets[target_id], default_cba=class_balancing_augmentation
    )
    new_augmentation = get_augmentation_rules(rules=asdict(class_balancing_augmentation), num_ir=num_ir)[0]
    new_augmentation.mixup = mixup
    target_augmentations.append(new_augmentation)
    return len(target_augmentations) - 1, target_augmentations


def get_class_balancing_augmentation(target: TargetFile, default_cba: AugmentationRule) -> AugmentationRule:
    """Get the class balancing augmentation rule for the given target"""
    if target.class_balancing_augmentation is not None:
        return target.class_balancing_augmentation
    return default_cba



# ruff: noqa: S608
from functools import cached_property
from functools import lru_cache
from functools import partial
from sqlite3 import Connection
from sqlite3 import Cursor
from typing import Any

from .datatypes import ASRConfigs
from .datatypes import AudioF
from .datatypes import AudioT
from .datatypes import ClassCount
from .datatypes import Feature
from .datatypes import FeatureGeneratorConfig
from .datatypes import FeatureGeneratorInfo
from .datatypes import GeneralizedIDs
from .datatypes import ImpulseResponseFile
from .datatypes import MetricDoc
from .datatypes import MetricDocs
from .datatypes import Mixture
from .datatypes import NoiseFile
from .datatypes import Segsnr
from .datatypes import SpectralMask
from .datatypes import SpeechMetadata
from .datatypes import TargetFile
from .datatypes import TransformConfig
from .datatypes import TruthConfigs
from .datatypes import TruthDict
from .datatypes import UniversalSNR


def db_file(location: str, test: bool = False) -> str:
    from os.path import join

    if test:
        name = "mixdb_test.db"
    else:
        name = "mixdb.db"

    return join(location, name)


def db_connection(
    location: str,
    create: bool = False,
    readonly: bool = True,
    test: bool = False,
    verbose: bool = False,
) -> Connection:
    import sqlite3
    from os import remove
    from os.path import exists

    name = db_file(location, test)
    if create and exists(name):
        remove(name)

    if not create and not exists(name):
        raise OSError(f"Could not find mixture database in {location}")

    if not create and readonly:
        name += "?mode=ro"

    connection = sqlite3.connect("file:" + name, uri=True, timeout=20)

    if verbose:
        connection.set_trace_callback(print)

    return connection


class SQLiteContextManager:
    def __init__(self, location: str, test: bool = False) -> None:
        self.location = location
        self.test = test

    def __enter__(self) -> Cursor:
        self.con = db_connection(location=self.location, test=self.test)
        self.cur = self.con.cursor()
        return self.cur

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.con.close()


class MixtureDatabase:
    def __init__(self, location: str, test: bool = False, use_cache: bool = True) -> None:
        import json
        from os.path import exists

        from .config import load_config

        self.location = location
        self.test = test
        self.use_cache = use_cache

        if not exists(db_file(self.location, self.test)):
            raise OSError(f"Could not find mixture database in {self.location}")

        self.db = partial(SQLiteContextManager, self.location, self.test)

        # Check config.yml to see if asr_configs has changed and update database if needed
        config = load_config(self.location)
        new_asr_configs = json.dumps(config["asr_configs"])
        with self.db() as c:
            old_asr_configs = c.execute("SELECT top.asr_configs FROM top").fetchone()

        if old_asr_configs is not None and new_asr_configs != old_asr_configs[0]:
            con = db_connection(location=self.location, readonly=False, test=self.test)
            con.execute("UPDATE top SET asr_configs = ? WHERE ? = id", (new_asr_configs,))
            con.commit()
            con.close()

    @cached_property
    def json(self) -> str:
        from .datatypes import MixtureDatabaseConfig

        config = MixtureDatabaseConfig(
            asr_configs=self.asr_configs,
            class_balancing=self.class_balancing,
            class_labels=self.class_labels,
            class_weights_threshold=self.class_weights_thresholds,
            feature=self.feature,
            impulse_response_files=self.impulse_response_files,
            mixtures=self.mixtures(),
            noise_mix_mode=self.noise_mix_mode,
            noise_files=self.noise_files,
            num_classes=self.num_classes,
            spectral_masks=self.spectral_masks,
            target_files=self.target_files,
        )
        return config.to_json(indent=2)

    def save(self) -> None:
        """Save the MixtureDatabase as a JSON file"""
        from os.path import join

        json_name = join(self.location, "mixdb.json")
        with open(file=json_name, mode="w") as file:
            file.write(self.json)

    @cached_property
    def fg_config(self) -> FeatureGeneratorConfig:
        return FeatureGeneratorConfig(
            feature_mode=self.feature,
            truth_parameters=self.truth_parameters,
        )

    @cached_property
    def fg_info(self) -> FeatureGeneratorInfo:
        from .helpers import get_feature_generator_info

        return get_feature_generator_info(self.fg_config)

    @cached_property
    def truth_parameters(self) -> dict[str, int | None]:
        with self.db() as c:
            rows = c.execute("SELECT * FROM truth_parameters").fetchall()
            truth_parameters: dict[str, int | None] = {}
            for row in rows:
                truth_parameters[row[1]] = row[2]
            return truth_parameters

    @cached_property
    def num_classes(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT top.num_classes FROM top").fetchone()[0])

    @cached_property
    def noise_mix_mode(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT top.noise_mix_mode FROM top").fetchone()[0])

    @cached_property
    def asr_configs(self) -> ASRConfigs:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT top.asr_configs FROM top").fetchone()[0])

    @cached_property
    def supported_metrics(self) -> MetricDocs:
        metrics = MetricDocs(
            [
                MetricDoc("Mixture Metrics", "mxsnr", "SNR specification in dB"),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnr_avg",
                    "Segmental SNR average over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnr_std",
                    "Segmental SNR standard deviation over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdb_avg",
                    "Segmental SNR average of the dB frame values over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdb_std",
                    "Segmental SNR standard deviation of the dB frame values over all frames",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrf_avg",
                    "Per-bin segmental SNR average over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrf_std",
                    "Per-bin segmental SNR standard deviation over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdbf_avg",
                    "Per-bin segmental average of the dB frame values over all frames (using feature transform)",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxssnrdbf_std",
                    "Per-bin segmental standard deviation of the dB frame values over all frames (using feature transform)",
                ),
                MetricDoc("Mixture Metrics", "mxpesq", "PESQ of mixture versus true targets"),
                MetricDoc(
                    "Mixture Metrics",
                    "mxwsdr",
                    "Weighted signal distortion ratio of mixture versus true targets",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxpd",
                    "Phase distance between mixture and true targets",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxstoi",
                    "Short term objective intelligibility of mixture versus true targets",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcsig",
                    "Predicted rating of speech distortion of mixture versus true targets",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcbak",
                    "Predicted rating of background distortion of mixture versus true targets",
                ),
                MetricDoc(
                    "Mixture Metrics",
                    "mxcovl",
                    "Predicted rating of overall quality of mixture versus true targets",
                ),
                MetricDoc("Mixture Metrics", "ssnr", "Segmental SNR"),
                MetricDoc("Mixture Metrics", "mxtdco", "Mixture target DC offset"),
                MetricDoc("Mixture Metrics", "mxtmin", "Mixture target min level"),
                MetricDoc("Mixture Metrics", "mxtmax", "Mixture target max levl"),
                MetricDoc("Mixture Metrics", "mxtpkdb", "Mixture target Pk lev dB"),
                MetricDoc("Mixture Metrics", "mxtlrms", "Mixture target RMS lev dB"),
                MetricDoc("Mixture Metrics", "mxtpkr", "Mixture target RMS Pk dB"),
                MetricDoc("Mixture Metrics", "mxttr", "Mixture target RMS Tr dB"),
                MetricDoc("Mixture Metrics", "mxtcr", "Mixture target Crest factor"),
                MetricDoc("Mixture Metrics", "mxtfl", "Mixture target Flat factor"),
                MetricDoc("Mixture Metrics", "mxtpkc", "Mixture target Pk count"),
                MetricDoc("Targets Metrics", "tdco", "Targets DC offset"),
                MetricDoc("Targets Metrics", "tmin", "Targets min level"),
                MetricDoc("Targets Metrics", "tmax", "Targets max levl"),
                MetricDoc("Targets Metrics", "tpkdb", "Targets Pk lev dB"),
                MetricDoc("Targets Metrics", "tlrms", "Targets RMS lev dB"),
                MetricDoc("Targets Metrics", "tpkr", "Targets RMS Pk dB"),
                MetricDoc("Targets Metrics", "ttr", "Targets RMS Tr dB"),
                MetricDoc("Targets Metrics", "tcr", "Targets Crest factor"),
                MetricDoc("Targets Metrics", "tfl", "Targets Flat factor"),
                MetricDoc("Targets Metrics", "tpkc", "Targets Pk count"),
                MetricDoc("Noise Metrics", "ndco", "Noise DC offset"),
                MetricDoc("Noise Metrics", "nmin", "Noise min level"),
                MetricDoc("Noise Metrics", "nmax", "Noise max levl"),
                MetricDoc("Noise Metrics", "npkdb", "Noise Pk lev dB"),
                MetricDoc("Noise Metrics", "nlrms", "Noise RMS lev dB"),
                MetricDoc("Noise Metrics", "npkr", "Noise RMS Pk dB"),
                MetricDoc("Noise Metrics", "ntr", "Noise RMS Tr dB"),
                MetricDoc("Noise Metrics", "ncr", "Noise Crest factor"),
                MetricDoc("Noise Metrics", "nfl", "Noise Flat factor"),
                MetricDoc("Noise Metrics", "npkc", "Noise Pk count"),
                MetricDoc(
                    "Truth Metrics",
                    "sedavg",
                    "(not implemented) Average SED activity over all frames [truth_parameters, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedcnt",
                    "(not implemented) Count in number of frames that SED is active [truth_parameters, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedtop3",
                    "(not implemented) 3 most active by largest sedavg [3, 1]",
                ),
                MetricDoc(
                    "Truth Metrics",
                    "sedtopn",
                    "(not implemented) N most active by largest sedavg [N, 1]",
                ),
            ]
        )
        for name in self.asr_configs:
            metrics.append(
                MetricDoc(
                    "Target Metrics",
                    f"mxtasr.{name}",
                    f"Mixture Target ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Target Metrics",
                    f"tasr.{name}",
                    f"Targets ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Mixture Metrics",
                    f"mxasr.{name}",
                    f"ASR text using {name} ASR as defined in mixdb asr_configs parameter",
                )
            )
            metrics.append(
                MetricDoc(
                    "Target Metrics",
                    f"basewer.{name}",
                    f"Word error rate of tasr.{name} vs. speech text metadata for the target",
                )
            )
            metrics.append(
                MetricDoc(
                    "Mixture Metrics",
                    f"mxwer.{name}",
                    f"Word error rate of mxasr.{name} vs. tasr.{name}",
                )
            )

        return metrics

    @cached_property
    def class_balancing(self) -> bool:
        with self.db() as c:
            return bool(c.execute("SELECT top.class_balancing FROM top").fetchone()[0])

    @cached_property
    def feature(self) -> str:
        with self.db() as c:
            return str(c.execute("SELECT top.feature FROM top").fetchone()[0])

    @cached_property
    def fg_decimation(self) -> int:
        return self.fg_info.decimation

    @cached_property
    def fg_stride(self) -> int:
        return self.fg_info.stride

    @cached_property
    def fg_step(self) -> int:
        return self.fg_info.step

    @cached_property
    def feature_parameters(self) -> int:
        return self.fg_info.feature_parameters

    @cached_property
    def ft_config(self) -> TransformConfig:
        return self.fg_info.ft_config

    @cached_property
    def eft_config(self) -> TransformConfig:
        return self.fg_info.eft_config

    @cached_property
    def it_config(self) -> TransformConfig:
        return self.fg_info.it_config

    @cached_property
    def transform_frame_ms(self) -> float:
        from .constants import SAMPLE_RATE

        return float(self.ft_config.overlap) / float(SAMPLE_RATE / 1000)

    @cached_property
    def feature_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_samples(self) -> int:
        return self.ft_config.overlap * self.fg_decimation * self.fg_stride

    @cached_property
    def feature_step_ms(self) -> float:
        return self.transform_frame_ms * self.fg_decimation * self.fg_step

    @cached_property
    def feature_step_samples(self) -> int:
        return self.ft_config.overlap * self.fg_decimation * self.fg_step

    def total_samples(self, m_ids: GeneralizedIDs = "*") -> int:
        samples = 0
        for m_id in self.mixids_to_list(m_ids):
            s = self.mixture(m_id).samples
            if s is not None:
                samples += s
        return samples

    def total_transform_frames(self, m_ids: GeneralizedIDs = "*") -> int:
        return self.total_samples(m_ids) // self.ft_config.overlap

    def total_feature_frames(self, m_ids: GeneralizedIDs = "*") -> int:
        return self.total_samples(m_ids) // self.feature_step_samples

    def mixture_transform_frames(self, m_id: int) -> int:
        from .helpers import frames_from_samples

        return frames_from_samples(self.mixture(m_id).samples, self.ft_config.overlap)

    def mixture_feature_frames(self, m_id: int) -> int:
        from .helpers import frames_from_samples

        return frames_from_samples(self.mixture(m_id).samples, self.feature_step_samples)

    def mixids_to_list(self, m_ids: GeneralizedIDs = "*") -> list[int]:
        """Resolve generalized mixture IDs to a list of integers

        :param m_ids: Generalized mixture IDs
        :return: List of mixture ID integers
        """
        from .helpers import generic_ids_to_list

        return generic_ids_to_list(self.num_mixtures, m_ids)

    @cached_property
    def class_labels(self) -> list[str]:
        """Get class labels from db

        :return: Class labels
        """
        with self.db() as c:
            return [
                str(item[0])
                for item in c.execute("SELECT class_label.label FROM class_label ORDER BY class_label.id").fetchall()
            ]

    @cached_property
    def class_weights_thresholds(self) -> list[float]:
        """Get class weights thresholds from db

        :return: Class weights thresholds
        """
        with self.db() as c:
            return [
                float(item[0])
                for item in c.execute(
                    "SELECT class_weights_threshold.threshold FROM class_weights_threshold"
                ).fetchall()
            ]

    @cached_property
    def truth_configs(self) -> TruthConfigs:
        """Get truth configs from db

        :return: Truth configs
        """
        import json

        from .datatypes import TruthConfig

        with self.db() as c:
            truth_configs: TruthConfigs = {}
            for truth_config_record in c.execute("SELECT truth_config.config FROM truth_config").fetchall():
                truth_config = json.loads(truth_config_record[0])
                if truth_config["name"] not in truth_configs:
                    truth_configs[truth_config["name"]] = TruthConfig(
                        function=truth_config["function"],
                        stride_reduction=truth_config["stride_reduction"],
                        config=truth_config["config"],
                    )
            return truth_configs

    def target_truth_configs(self, t_id: int) -> TruthConfigs:
        return _target_truth_configs(self.db, t_id, self.use_cache)

    @cached_property
    def random_snrs(self) -> list[float]:
        """Get random snrs from db

        :return: Random SNRs
        """
        with self.db() as c:
            return list(
                {
                    float(item[0])
                    for item in c.execute("SELECT mixture.snr FROM mixture WHERE mixture.random_snr == 1").fetchall()
                }
            )

    @cached_property
    def snrs(self) -> list[float]:
        """Get snrs from db

        :return: SNRs
        """
        with self.db() as c:
            return list(
                {
                    float(item[0])
                    for item in c.execute("SELECT mixture.snr FROM mixture WHERE mixture.random_snr == 0").fetchall()
                }
            )

    @cached_property
    def all_snrs(self) -> list[UniversalSNR]:
        return sorted(
            set(
                [UniversalSNR(is_random=False, value=snr) for snr in self.snrs]
                + [UniversalSNR(is_random=True, value=snr) for snr in self.random_snrs]
            )
        )

    @cached_property
    def spectral_masks(self) -> list[SpectralMask]:
        """Get spectral masks from db

        :return: Spectral masks
        """
        from .db_datatypes import SpectralMaskRecord

        with self.db() as c:
            spectral_masks = [
                SpectralMaskRecord(*result) for result in c.execute("SELECT * FROM spectral_mask").fetchall()
            ]
            return [
                SpectralMask(
                    f_max_width=spectral_mask.f_max_width,
                    f_num=spectral_mask.f_num,
                    t_max_width=spectral_mask.t_max_width,
                    t_num=spectral_mask.t_num,
                    t_max_percent=spectral_mask.t_max_percent,
                )
                for spectral_mask in spectral_masks
            ]

    def spectral_mask(self, sm_id: int) -> SpectralMask:
        """Get spectral mask with ID from db

        :param sm_id: Spectral mask ID
        :return: Spectral mask
        """
        return _spectral_mask(self.db, sm_id, self.use_cache)

    @cached_property
    def target_files(self) -> list[TargetFile]:
        """Get target files from db

        :return: Target files
        """
        import json

        from .datatypes import TruthConfig
        from .datatypes import TruthConfigs
        from .db_datatypes import TargetFileRecord

        with self.db() as c:
            target_files: list[TargetFile] = []
            target_file_records = [
                TargetFileRecord(*result) for result in c.execute("SELECT * FROM target_file").fetchall()
            ]
            for target_file_record in target_file_records:
                truth_configs: TruthConfigs = {}
                for truth_config_records in c.execute(
                    """
                    SELECT truth_config.config
                    FROM truth_config, target_file_truth_config
                    WHERE ? = target_file_truth_config.target_file_id
                    AND truth_config.id = target_file_truth_config.truth_config_id
                    """,
                    (target_file_record.id,),
                ).fetchall():
                    truth_config = json.loads(truth_config_records[0])
                    truth_configs[truth_config["name"]] = TruthConfig(
                        function=truth_config["function"],
                        stride_reduction=truth_config["stride_reduction"],
                        config=truth_config["config"],
                    )
                target_files.append(
                    TargetFile(
                        name=target_file_record.name,
                        samples=target_file_record.samples,
                        class_indices=json.loads(target_file_record.class_indices),
                        level_type=target_file_record.level_type,
                        truth_configs=truth_configs,
                        speaker_id=target_file_record.speaker_id,
                    )
                )
            return target_files

    @cached_property
    def target_file_ids(self) -> list[int]:
        """Get target file IDs from db

        :return: List of target file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in c.execute("SELECT target_file.id FROM target_file").fetchall()]

    def target_file(self, t_id: int) -> TargetFile:
        """Get target file with ID from db

        :param t_id: Target file ID
        :return: Target file
        """
        return _target_file(self.db, t_id, self.use_cache)

    @cached_property
    def num_target_files(self) -> int:
        """Get number of target files from db

        :return: Number of target files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(target_file.id) FROM target_file").fetchone()[0])

    @cached_property
    def noise_files(self) -> list[NoiseFile]:
        """Get noise files from db

        :return: Noise files
        """
        with self.db() as c:
            return [
                NoiseFile(name=noise[0], samples=noise[1])
                for noise in c.execute("SELECT noise_file.name, samples FROM noise_file").fetchall()
            ]

    @cached_property
    def noise_file_ids(self) -> list[int]:
        """Get noise file IDs from db

        :return: List of noise file IDs
        """
        with self.db() as c:
            return [int(item[0]) for item in c.execute("SELECT noise_file.id FROM noise_file").fetchall()]

    def noise_file(self, n_id: int) -> NoiseFile:
        """Get noise file with ID from db

        :param n_id: Noise file ID
        :return: Noise file
        """
        return _noise_file(self.db, n_id, self.use_cache)

    @cached_property
    def num_noise_files(self) -> int:
        """Get number of noise files from db

        :return: Number of noise files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(noise_file.id) FROM noise_file").fetchone()[0])

    @cached_property
    def impulse_response_files(self) -> list[ImpulseResponseFile]:
        """Get impulse response files from db

        :return: Impulse response files
        """
        import json

        from .datatypes import ImpulseResponseFile

        with self.db() as c:
            return [
                ImpulseResponseFile(impulse_response[1], json.loads(impulse_response[2]))
                for impulse_response in c.execute(
                    "SELECT impulse_response_file.* FROM impulse_response_file"
                ).fetchall()
            ]

    @cached_property
    def impulse_response_file_ids(self) -> list[int]:
        """Get impulse response file IDs from db

        :return: List of impulse response file IDs
        """
        with self.db() as c:
            return [
                int(item[0])
                for item in c.execute("SELECT impulse_response_file.id FROM impulse_response_file").fetchall()
            ]

    def impulse_response_file(self, ir_id: int | None) -> str | None:
        """Get impulse response file with ID from db

        :param ir_id: Impulse response file ID
        :return: Noise
        """
        if ir_id is None:
            return None
        return _impulse_response_file(self.db, ir_id, self.use_cache)

    @cached_property
    def num_impulse_response_files(self) -> int:
        """Get number of impulse response files from db

        :return: Number of impulse response files
        """
        with self.db() as c:
            return int(c.execute("SELECT count(impulse_response_file.id) FROM impulse_response_file").fetchone()[0])

    def mixtures(self) -> list[Mixture]:
        """Get mixtures from db

        :return: Mixtures
        """
        from .db_datatypes import MixtureRecord
        from .db_datatypes import TargetRecord
        from .helpers import to_mixture
        from .helpers import to_target

        with self.db() as c:
            mixtures: list[Mixture] = []
            for mixture in [MixtureRecord(*record) for record in c.execute("SELECT * FROM mixture").fetchall()]:
                targets = [
                    to_target(TargetRecord(*target))
                    for target in c.execute(
                        """
                        SELECT target.*
                        FROM target, mixture_target
                        WHERE ? = mixture_target.mixture_id AND target.id = mixture_target.target_id
                        """,
                        (mixture.id,),
                    ).fetchall()
                ]
                mixtures.append(to_mixture(mixture, targets))
            return mixtures

    @cached_property
    def mixture_ids(self) -> list[int]:
        """Get mixture IDs from db

        :return: List of zero-based mixture IDs
        """
        with self.db() as c:
            return [int(item[0]) - 1 for item in c.execute("SELECT mixture.id FROM mixture").fetchall()]

    def mixture(self, m_id: int) -> Mixture:
        """Get mixture record with ID from db

        :param m_id: Zero-based mixture ID
        :return: Mixture record
        """
        return _mixture(self.db, m_id, self.use_cache)

    @cached_property
    def mixid_width(self) -> int:
        with self.db() as c:
            return int(c.execute("SELECT top.mixid_width FROM top").fetchone()[0])

    def mixture_location(self, m_id: int) -> str:
        """Get the file location for the give mixture ID

        :param m_id: Zero-based mixture ID
        :return: File location
        """
        from os.path import join

        return join(self.location, self.mixture(m_id).name)

    @cached_property
    def num_mixtures(self) -> int:
        """Get number of mixtures from db

        :return: Number of mixtures
        """
        with self.db() as c:
            return int(c.execute("SELECT count(mixture.id) FROM mixture").fetchone()[0])

    def read_mixture_data(self, m_id: int, items: list[str] | str) -> Any:
        """Read mixture data

        :param m_id: Zero-based mixture ID
        :param items: String(s) of dataset(s) to retrieve
        :return: Data (or tuple of data)
        """
        from sonusai.mixture import read_cached_data

        return read_cached_data(self.location, "mixture", self.mixture(m_id).name, items)

    def read_target_audio(self, t_id: int) -> AudioT:
        """Read target audio

        :param t_id: Target ID
        :return: Target audio
        """
        from .audio import read_audio

        return read_audio(self.target_file(t_id).name, self.use_cache)

    def augmented_noise_audio(self, mixture: Mixture) -> AudioT:
        """Get augmented noise audio

        :param mixture: Mixture
        :return: Augmented noise audio
        """
        from .audio import read_audio
        from .audio import read_ir
        from .augmentation import apply_augmentation
        from .augmentation import apply_impulse_response

        noise = self.noise_file(mixture.noise.file_id)
        audio = read_audio(noise.name, self.use_cache)
        audio = apply_augmentation(audio, mixture.noise.augmentation)
        if mixture.noise.augmentation.ir is not None:
            audio = apply_impulse_response(
                audio,
                read_ir(self.impulse_response_file(mixture.noise.augmentation.ir), self.use_cache),  # pyright: ignore [reportArgumentType]
            )

        return audio

    def mixture_class_indices(self, m_id: int) -> list[int]:
        class_indices: list[int] = []
        for t_id in self.mixture(m_id).target_ids:
            class_indices.extend(self.target_file(t_id).class_indices)
        return sorted(set(class_indices))

    def mixture_targets(self, m_id: int, force: bool = False) -> list[AudioT]:
        """Get the list of augmented target audio data (one per target in the mixup) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: List of augmented target audio data (one per target in the mixup)
        """
        from .augmentation import apply_augmentation
        from .augmentation import apply_gain
        from .augmentation import pad_audio_to_length

        if not force:
            targets_audio = self.read_mixture_data(m_id, "targets")
            if targets_audio is not None:
                return list(targets_audio)

        mixture = self.mixture(m_id)
        if mixture is None:
            raise ValueError(f"Could not find mixture for m_id: {m_id}")

        targets_audio = []
        for target in mixture.targets:
            target_audio = self.read_target_audio(target.file_id)
            target_audio = apply_augmentation(
                audio=target_audio,
                augmentation=target.augmentation,
                frame_length=self.feature_step_samples,
            )
            target_audio = apply_gain(audio=target_audio, gain=mixture.target_snr_gain)
            target_audio = pad_audio_to_length(audio=target_audio, length=mixture.samples)
            targets_audio.append(target_audio)

        return targets_audio

    def mixture_targets_f(self, m_id: int, targets: list[AudioT] | None = None, force: bool = False) -> list[AudioF]:
        """Get the list of augmented target transform data (one per target in the mixup) for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: List of augmented target transform data (one per target in the mixup)
        """
        from .helpers import forward_transform

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        return [forward_transform(target, self.ft_config) for target in targets]

    def mixture_target(self, m_id: int, targets: list[AudioT] | None = None, force: bool = False) -> AudioT:
        """Get the augmented target audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented target audio data
        """
        from .helpers import get_target

        if not force:
            target = self.read_mixture_data(m_id, "target")
            if target is not None:
                return target

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        return get_target(self, self.mixture(m_id), targets)

    def mixture_target_f(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        force: bool = False,
    ) -> AudioF:
        """Get the augmented target transform data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio for the given m_id
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented target transform data
        """
        from .helpers import forward_transform

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        return forward_transform(target, self.ft_config)

    def mixture_noise(self, m_id: int, force: bool = False) -> AudioT:
        """Get the augmented noise audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented noise audio data
        """
        from .audio import get_next_noise
        from .augmentation import apply_gain

        if not force:
            noise = self.read_mixture_data(m_id, "noise")
            if noise is not None:
                return noise

        mixture = self.mixture(m_id)
        noise = self.augmented_noise_audio(mixture)
        noise = get_next_noise(audio=noise, offset=mixture.noise_offset, length=mixture.samples)
        return apply_gain(audio=noise, gain=mixture.noise_snr_gain)

    def mixture_noise_f(self, m_id: int, noise: AudioT | None = None, force: bool = False) -> AudioF:
        """Get the augmented noise transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Augmented noise transform data
        """
        from .helpers import forward_transform

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        return forward_transform(noise, self.ft_config)

    def mixture_mixture(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
    ) -> AudioT:
        """Get the mixture audio data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture audio data
        """
        if not force:
            mixture = self.read_mixture_data(m_id, "mixture")
            if mixture is not None:
                return mixture

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        return target + noise

    def mixture_mixture_f(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        force: bool = False,
    ) -> AudioF:
        """Get the mixture transform for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Mixture transform data
        """
        from .helpers import forward_transform
        from .spectral_mask import apply_spectral_mask

        if force or mixture is None:
            mixture = self.mixture_mixture(m_id, targets, target, noise, force)

        mixture_f = forward_transform(mixture, self.ft_config)

        m = self.mixture(m_id)
        if m.spectral_mask_id is not None:
            mixture_f = apply_spectral_mask(
                audio_f=mixture_f,
                spectral_mask=self.spectral_mask(int(m.spectral_mask_id)),
                seed=m.spectral_mask_seed,
            )

        return mixture_f

    def mixture_truth_t(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        force: bool = False,
    ) -> list[TruthDict]:
        """Get the truth_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup) for the given mixture ID
        :param noise: Augmented noise audio data for the given mixture ID
        :param mixture: Mixture audio data for the given mixture ID
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: list of truth_t data
        """
        from .truth import truth_function

        if not force:
            truth_t = self.read_mixture_data(m_id, "truth_t")
            if truth_t is not None:
                return truth_t

        if force or targets is None:
            targets = self.mixture_targets(m_id, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        if force or mixture is None:
            mixture = self.mixture_mixture(m_id, targets=targets, noise=noise, force=force)

        if not all(len(target) == self.mixture(m_id).samples for target in targets):
            raise ValueError("Lengths of targets do not match length of mixture")

        if len(noise) != self.mixture(m_id).samples:
            raise ValueError("Length of noise does not match length of mixture")

        return truth_function(self, m_id)

    def mixture_segsnr_t(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
    ) -> Segsnr:
        """Get the segsnr_t data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: segsnr_t data
        """
        import numpy as np
        import torch
        from pyaaware import ForwardTransform

        if not force:
            segsnr_t = self.read_mixture_data(m_id, "segsnr_t")
            if segsnr_t is not None:
                return segsnr_t

        if force or target is None:
            target = self.mixture_target(m_id, targets, force)

        if force or noise is None:
            noise = self.mixture_noise(m_id, force)

        ft = ForwardTransform(
            length=self.ft_config.length,
            overlap=self.ft_config.overlap,
            bin_start=self.ft_config.bin_start,
            bin_end=self.ft_config.bin_end,
            ttype=self.ft_config.ttype,
        )

        mixture = self.mixture(m_id)

        segsnr_t = np.empty(mixture.samples, dtype=np.float32)

        target_energy = ft.execute_all(torch.from_numpy(target))[1].numpy()
        noise_energy = ft.execute_all(torch.from_numpy(noise))[1].numpy()

        offsets = range(0, mixture.samples, self.ft_config.overlap)
        if len(target_energy) != len(offsets):
            raise ValueError(
                f"Number of frames in energy, {len(target_energy)},"
                f" is not number of frames in mixture, {len(offsets)}"
            )

        for idx, offset in enumerate(offsets):
            indices = slice(offset, offset + self.ft_config.overlap)

            if noise_energy[idx] == 0:
                snr = np.float32(np.inf)
            else:
                snr = np.float32(target_energy[idx] / noise_energy[idx])

            segsnr_t[indices] = snr

        return segsnr_t

    def mixture_segsnr(
        self,
        m_id: int,
        segsnr_t: Segsnr | None = None,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        noise: AudioT | None = None,
        force: bool = False,
    ) -> Segsnr:
        """Get the segsnr data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param segsnr_t: segsnr_t data
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: segsnr data
        """
        if not force:
            segsnr = self.read_mixture_data(m_id, "segsnr")
            if segsnr is not None:
                return segsnr

            segsnr_t = self.read_mixture_data(m_id, "segsnr_t")
            if segsnr_t is not None:
                return segsnr_t[0 :: self.ft_config.overlap]

        if force or segsnr_t is None:
            segsnr_t = self.mixture_segsnr_t(m_id, targets, target, noise, force)

        return segsnr_t[0 :: self.ft_config.overlap]

    def mixture_ft(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        target: AudioT | None = None,
        noise: AudioT | None = None,
        mixture_f: AudioF | None = None,
        mixture: AudioT | None = None,
        truth_t: list[TruthDict] | None = None,
        force: bool = False,
    ) -> tuple[Feature, TruthDict]:
        """Get the feature and truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param target: Augmented target audio data
        :param noise: Augmented noise audio data
        :param mixture_f: Mixture transform data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Tuple of (feature, truth_f) data
        """
        from pyaaware import FeatureGenerator

        from .truth import truth_stride_reduction

        if not force:
            feature, truth_f = self.read_mixture_data(m_id, ["feature", "truth_f"])
            if feature is not None and truth_f is not None:
                return feature, truth_f

        if force or mixture_f is None:
            mixture_f = self.mixture_mixture_f(
                m_id=m_id,
                targets=targets,
                target=target,
                noise=noise,
                mixture=mixture,
                force=force,
            )

        if force or truth_t is None:
            truth_t = self.mixture_truth_t(m_id=m_id, targets=targets, noise=noise, force=force)

        fg = FeatureGenerator(self.fg_config.feature_mode, self.fg_config.truth_parameters)

        # TODO: handle mixup in truth_t
        feature, truth_f = fg.execute_all(mixture_f, truth_t[0])
        if truth_f is not None:
            for key in self.truth_configs:
                if self.truth_parameters[key] is not None:
                    truth_f[key] = truth_stride_reduction(truth_f[key], self.truth_configs[key].stride_reduction)
        else:
            raise TypeError("Unexpected truth of None from feature generator")

        return feature, truth_f

    def mixture_feature(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        truth_t: list[TruthDict] | None = None,
        force: bool = False,
    ) -> Feature:
        """Get the feature data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: Feature data
        """
        feature, _ = self.mixture_ft(
            m_id=m_id,
            targets=targets,
            noise=noise,
            mixture=mixture,
            truth_t=truth_t,
            force=force,
        )
        return feature

    def mixture_truth_f(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        noise: AudioT | None = None,
        mixture: AudioT | None = None,
        truth_t: list[TruthDict] | None = None,
        force: bool = False,
    ) -> TruthDict:
        """Get the truth_f data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio data (one per target in the mixup)
        :param noise: Augmented noise audio data
        :param mixture: Mixture audio data
        :param truth_t: truth_t
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: truth_f data
        """
        _, truth_f = self.mixture_ft(
            m_id=m_id,
            targets=targets,
            noise=noise,
            mixture=mixture,
            truth_t=truth_t,
            force=force,
        )
        return truth_f

    def mixture_class_count(
        self,
        m_id: int,
        targets: list[AudioT] | None = None,
        noise: AudioT | None = None,
        truth_t: list[TruthDict] | None = None,
    ) -> ClassCount:
        """Compute the number of frames for which each class index is active for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param targets: List of augmented target audio (one per target in the mixup)
        :param noise: Augmented noise audio
        :param truth_t: truth_t
        :return: List of class counts
        """
        import numpy as np

        if truth_t is None:
            truth_t = self.mixture_truth_t(m_id, targets, noise)

        class_count = [0] * self.num_classes
        num_classes = self.num_classes
        if "sed" in self.truth_configs:
            for cl in range(num_classes):
                # TODO: handle mixup in truth_t
                class_count[cl] = int(np.sum(truth_t[0]["sed"][:, cl] >= self.class_weights_thresholds[cl]))

        return class_count

    @cached_property
    def speaker_metadata_tiers(self) -> list[str]:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT speaker_metadata_tiers FROM top WHERE 1 = id").fetchone()[0])

    @cached_property
    def textgrid_metadata_tiers(self) -> list[str]:
        import json

        with self.db() as c:
            return json.loads(c.execute("SELECT textgrid_metadata_tiers FROM top WHERE 1 = id").fetchone()[0])

    @cached_property
    def speech_metadata_tiers(self) -> list[str]:
        return sorted(set(self.speaker_metadata_tiers + self.textgrid_metadata_tiers))

    def speaker(self, s_id: int | None, tier: str) -> str | None:
        return _speaker(self.db, s_id, tier, self.use_cache)

    def speech_metadata(self, tier: str) -> list[str]:
        from .helpers import get_textgrid_tier_from_target_file

        results: set[str] = set()
        if tier in self.textgrid_metadata_tiers:
            for target_file in self.target_files:
                data = get_textgrid_tier_from_target_file(target_file.name, tier)
                if data is None:
                    continue
                if isinstance(data, list):
                    for item in data:
                        results.add(item.label)
                else:
                    results.add(data)
        elif tier in self.speaker_metadata_tiers:
            for target_file in self.target_files:
                data = self.speaker(target_file.speaker_id, tier)
                if data is not None:
                    results.add(data)

        return sorted(results)

    def mixture_speech_metadata(self, mixid: int, tier: str) -> list[SpeechMetadata]:
        from praatio.utilities.constants import Interval

        from .helpers import get_textgrid_tier_from_target_file

        results: list[SpeechMetadata] = []
        is_textgrid = tier in self.textgrid_metadata_tiers
        if is_textgrid:
            for target in self.mixture(mixid).targets:
                data = get_textgrid_tier_from_target_file(self.target_file(target.file_id).name, tier)
                if isinstance(data, list):
                    # Check for tempo augmentation and adjust Interval start and end data as needed
                    entries = []
                    for entry in data:
                        if target.augmentation.tempo is not None:
                            entries.append(
                                Interval(
                                    entry.start / target.augmentation.tempo,
                                    entry.end / target.augmentation.tempo,
                                    entry.label,
                                )
                            )
                        else:
                            entries.append(entry)
                    results.append(entries)
                else:
                    results.append(data)
        else:
            for target in self.mixture(mixid).targets:
                results.append(self.speaker(self.target_file(target.file_id).speaker_id, tier))

        return results

    def mixids_for_speech_metadata(
        self,
        tier: str | None = None,
        value: str | None = None,
        where: str | None = None,
    ) -> list[int]:
        """Get a list of mixture IDs for the given speech metadata tier.

        If 'where' is None, then include mixture IDs whose tier values are equal to the given 'value'.
        If 'where' is not None, then ignore 'value' and use the given SQL where clause to determine
        which entries to include.

        Examples:
        >>> mixdb = MixtureDatabase('/mixdb_location')

        >>> mixids = mixdb.mixids_for_speech_metadata('speaker_id', 'TIMIT_ABW0')
        Get mixture IDs for mixtures with speakers whose speaker_ids are 'TIMIT_ABW0'.

        >>> mixids = mixdb.mixids_for_speech_metadata(where='age >= 27')
        Get mixture IDs for mixtures with speakers whose ages are greater than or equal to 27.

        >>> mixids = mixdb.mixids_for_speech_metadata(where="dialect in ('New York City', 'Northern')")
        Get mixture IDs for mixtures with speakers whose dialects are either 'New York City' or 'Northern'.
        """
        if value is None and where is None:
            raise ValueError("Must provide either value or where")

        if where is None:
            if tier is None:
                raise ValueError("Must provide tier")
            where = f"{tier} = '{value}'"

        if tier is not None and tier in self.textgrid_metadata_tiers:
            raise ValueError(f"TextGrid tier data, '{tier}', is not supported in mixids_for_speech_metadata().")

        with self.db() as c:
            results = c.execute(f"SELECT id FROM speaker WHERE {where}").fetchall()
            speaker_ids = ",".join(map(str, [i[0] for i in results]))

            results = c.execute(f"SELECT id FROM target_file WHERE speaker_id IN ({speaker_ids})").fetchall()
            target_file_ids = ",".join(map(str, [i[0] for i in results]))

            results = c.execute(
                f"SELECT mixture_id FROM mixture_target WHERE mixture_target.target_id IN ({target_file_ids})"
            ).fetchall()

        return [mixture_id[0] - 1 for mixture_id in results]

    def mixture_all_speech_metadata(self, m_id: int) -> list[dict[str, SpeechMetadata]]:
        from .helpers import mixture_all_speech_metadata

        return mixture_all_speech_metadata(self, self.mixture(m_id))

    def cached_metrics(self, m_ids: GeneralizedIDs = "*") -> list[str]:
        """Get list of cached metrics for all mixtures."""
        from glob import glob
        from os.path import join
        from pathlib import Path

        supported_metrics = self.supported_metrics.names
        first = True
        result: set[str] = set()
        for m_id in self.mixids_to_list(m_ids):
            mixture_dir = join(self.location, "mixture", self.mixture(m_id).name)
            found = {Path(f).stem for f in glob(join(mixture_dir, "*.pkl"))}
            if first:
                first = False
                for f in found:
                    if f in supported_metrics:
                        result.add(f)
            else:
                result = result & found

        return sorted(result)

    def mixture_metrics(self, m_id: int, metrics: list[str], force: bool = False) -> dict[str, Any]:
        """Get metrics data for the given mixture ID

        :param m_id: Zero-based mixture ID
        :param metrics: List of metrics to get
        :param force: Force computing data from original sources regardless of whether cached data exists
        :return: List of metric data
        """
        from collections.abc import Callable

        import numpy as np
        from pystoi import stoi

        from sonusai.metrics import calc_audio_stats
        from sonusai.metrics import calc_phase_distance
        from sonusai.metrics import calc_segsnr_f
        from sonusai.metrics import calc_segsnr_f_bin
        from sonusai.metrics import calc_speech
        from sonusai.metrics import calc_wer
        from sonusai.metrics import calc_wsdr
        from sonusai.mixture import SAMPLE_RATE
        from sonusai.mixture import AudioStatsMetrics
        from sonusai.mixture import SpeechMetrics
        from sonusai.utils import calc_asr

        def create_targets_audio() -> Callable[[], list[AudioT]]:
            state: list[AudioT] | None = None

            def get() -> list[AudioT]:
                nonlocal state
                if state is None:
                    state = self.mixture_targets(m_id)
                return state

            return get

        targets_audio = create_targets_audio()

        def create_target_audio() -> Callable[[], AudioT]:
            state: AudioT | None = None

            def get() -> AudioT:
                nonlocal state
                if state is None:
                    state = self.mixture_target(m_id)
                return state

            return get

        target_audio = create_target_audio()

        def create_target_f() -> Callable[[], AudioF]:
            state: AudioF | None = None

            def get() -> AudioF:
                nonlocal state
                if state is None:
                    state = self.mixture_targets_f(m_id)[0]
                return state

            return get

        target_f = create_target_f()

        def create_noise_audio() -> Callable[[], AudioT]:
            state: AudioT | None = None

            def get() -> AudioT:
                nonlocal state
                if state is None:
                    state = self.mixture_noise(m_id)
                return state

            return get

        noise_audio = create_noise_audio()

        def create_noise_f() -> Callable[[], AudioF]:
            state: AudioF | None = None

            def get() -> AudioF:
                nonlocal state
                if state is None:
                    state = self.mixture_noise_f(m_id)
                return state

            return get

        noise_f = create_noise_f()

        def create_mixture_audio() -> Callable[[], AudioT]:
            state: AudioT | None = None

            def get() -> AudioT:
                nonlocal state
                if state is None:
                    state = self.mixture_mixture(m_id)
                return state

            return get

        mixture_audio = create_mixture_audio()

        def create_segsnr_f() -> Callable[[], Segsnr]:
            state: Segsnr | None = None

            def get() -> Segsnr:
                nonlocal state
                if state is None:
                    state = self.mixture_segsnr(m_id)
                return state

            return get

        segsnr_f = create_segsnr_f()

        def create_speech() -> Callable[[], list[SpeechMetrics]]:
            state: list[SpeechMetrics] | None = None

            def get() -> list[SpeechMetrics]:
                nonlocal state
                if state is None:
                    state = []
                    for audio in targets_audio():
                        state.append(calc_speech(hypothesis=mixture_audio(), reference=audio))
                return state

            return get

        speech = create_speech()

        def create_targets_stats() -> Callable[[], list[AudioStatsMetrics]]:
            state: list[AudioStatsMetrics] | None = None

            def get() -> list[AudioStatsMetrics]:
                nonlocal state
                if state is None:
                    state = []
                    for audio in targets_audio():
                        state.append(calc_audio_stats(audio, self.fg_info.ft_config.length / SAMPLE_RATE))
                return state

            return get

        targets_stats = create_targets_stats()

        def create_target_stats() -> Callable[[], AudioStatsMetrics]:
            state: AudioStatsMetrics | None = None

            def get() -> AudioStatsMetrics:
                nonlocal state
                if state is None:
                    state = calc_audio_stats(target_audio(), self.fg_info.ft_config.length / SAMPLE_RATE)
                return state

            return get

        target_stats = create_target_stats()

        def create_noise_stats() -> Callable[[], AudioStatsMetrics]:
            state: AudioStatsMetrics | None = None

            def get() -> AudioStatsMetrics:
                nonlocal state
                if state is None:
                    state = calc_audio_stats(noise_audio(), self.fg_info.ft_config.length / SAMPLE_RATE)
                return state

            return get

        noise_stats = create_noise_stats()

        def create_asr_config() -> Callable[[str], dict]:
            state: dict[str, dict] = {}

            def get(asr_name) -> dict:
                nonlocal state
                if asr_name not in state:
                    value = self.asr_configs.get(asr_name, None)
                    if value is None:
                        raise ValueError(f"Unrecognized ASR name: '{asr_name}'")
                    state[asr_name] = value
                return state[asr_name]

            return get

        asr_config = create_asr_config()

        def create_targets_asr() -> Callable[[str], list[str]]:
            state: dict[str, list[str]] = {}

            def get(asr_name) -> list[str]:
                nonlocal state
                if asr_name not in state:
                    state[asr_name] = []
                    for audio in targets_audio():
                        state[asr_name].append(calc_asr(audio, **asr_config(asr_name)).text)
                return state[asr_name]

            return get

        targets_asr = create_targets_asr()

        def create_target_asr() -> Callable[[str], str]:
            state: dict[str, str] = {}

            def get(asr_name) -> str:
                nonlocal state
                if asr_name not in state:
                    state[asr_name] = calc_asr(target_audio(), **asr_config(asr_name)).text
                return state[asr_name]

            return get

        target_asr = create_target_asr()

        def create_mixture_asr() -> Callable[[str], str]:
            state: dict[str, str] = {}

            def get(asr_name) -> str:
                nonlocal state
                if asr_name not in state:
                    state[asr_name] = calc_asr(mixture_audio(), **asr_config(asr_name)).text
                return state[asr_name]

            return get

        mixture_asr = create_mixture_asr()

        def get_asr_name(m: str) -> str:
            parts = m.split(".")
            if len(parts) != 2:
                raise ValueError(f"Unrecognized format: '{m}'; must be of the form: '<metric>.<name>'")
            asr_name = parts[1]
            return asr_name

        def calc(m: str) -> Any:
            if m == "mxsnr":
                return self.mixture(m_id).snr

            # Get cached data first, if exists
            if not force:
                value = self.read_mixture_data(m_id, m)
                if value is not None:
                    return value

            # Otherwise, generate data as needed
            if m.startswith("mxwer"):
                asr_name = get_asr_name(m)

                if self.mixture(m_id).is_noise_only:
                    # noise only, ignore/reset target asr
                    return float("nan")

                if target_asr(asr_name):
                    return calc_wer(mixture_asr(asr_name), target_asr(asr_name)).wer * 100

                # TODO: should this be NaN like above?
                return float(0)

            if m.startswith("basewer"):
                asr_name = get_asr_name(m)

                text = self.mixture_speech_metadata(m_id, "text")[0]
                if not isinstance(text, str):
                    # TODO: should this be NaN like above?
                    return [float(0)] * len(targets_audio())

                return [calc_wer(t, text).wer * 100 for t in targets_asr(asr_name)]

            if m.startswith("mxasr"):
                return mixture_asr(get_asr_name(m))

            if m == "mxssnr_avg":
                return calc_segsnr_f(segsnr_f()).avg

            if m == "mxssnr_std":
                return calc_segsnr_f(segsnr_f()).std

            if m == "mxssnrdb_avg":
                return calc_segsnr_f(segsnr_f()).db_avg

            if m == "mxssnrdb_std":
                return calc_segsnr_f(segsnr_f()).db_std

            if m == "mxssnrf_avg":
                return calc_segsnr_f_bin(target_f(), noise_f()).avg

            if m == "mxssnrf_std":
                return calc_segsnr_f_bin(target_f(), noise_f()).std

            if m == "mxssnrdbf_avg":
                return calc_segsnr_f_bin(target_f(), noise_f()).db_avg

            if m == "mxssnrdbf_std":
                return calc_segsnr_f_bin(target_f(), noise_f()).db_std

            if m == "mxpesq":
                if self.mixture(m_id).is_noise_only:
                    return [0] * len(speech())
                return [s.pesq for s in speech()]

            if m == "mxcsig":
                if self.mixture(m_id).is_noise_only:
                    return [0] * len(speech())
                return [s.csig for s in speech()]

            if m == "mxcbak":
                if self.mixture(m_id).is_noise_only:
                    return [0] * len(speech())
                return [s.cbak for s in speech()]

            if m == "mxcovl":
                if self.mixture(m_id).is_noise_only:
                    return [0] * len(speech())
                return [s.covl for s in speech()]

            if m == "mxwsdr":
                mixture = mixture_audio()[:, np.newaxis]
                target = target_audio()[:, np.newaxis]
                noise = noise_audio()[:, np.newaxis]
                return calc_wsdr(
                    hypothesis=np.concatenate((mixture, noise), axis=1),
                    reference=np.concatenate((target, noise), axis=1),
                    with_log=True,
                )[0]

            if m == "mxpd":
                mixture_f = self.mixture_mixture_f(m_id)
                return calc_phase_distance(hypothesis=mixture_f, reference=target_f())[0]

            if m == "mxstoi":
                return stoi(
                    x=target_audio(),
                    y=mixture_audio(),
                    fs_sig=SAMPLE_RATE,
                    extended=False,
                )

            if m == "mxtdco":
                return target_stats().dco

            if m == "mxtmin":
                return target_stats().min

            if m == "mxtmax":
                return target_stats().max

            if m == "mxtpkdb":
                return target_stats().pkdb

            if m == "mxtlrms":
                return target_stats().lrms

            if m == "mxtpkr":
                return target_stats().pkr

            if m == "mxttr":
                return target_stats().tr

            if m == "mxtcr":
                return target_stats().cr

            if m == "mxtfl":
                return target_stats().fl

            if m == "mxtpkc":
                return target_stats().pkc

            if m == "tdco":
                return [t.dco for t in targets_stats()]

            if m == "tmin":
                return [t.min for t in targets_stats()]

            if m == "tmax":
                return [t.max for t in targets_stats()]

            if m == "tpkdb":
                return [t.pkdb for t in targets_stats()]

            if m == "tlrms":
                return [t.lrms for t in targets_stats()]

            if m == "tpkr":
                return [t.pkr for t in targets_stats()]

            if m == "ttr":
                return [t.tr for t in targets_stats()]

            if m == "tcr":
                return [t.cr for t in targets_stats()]

            if m == "tfl":
                return [t.fl for t in targets_stats()]

            if m == "tpkc":
                return [t.pkc for t in targets_stats()]

            if m.startswith("tasr"):
                return targets_asr(get_asr_name(m))

            if m.startswith("mxtasr"):
                return target_asr(get_asr_name(m))

            if m == "ndco":
                return noise_stats().dco

            if m == "nmin":
                return noise_stats().min

            if m == "nmax":
                return noise_stats().max

            if m == "npkdb":
                return noise_stats().pkdb

            if m == "nlrms":
                return noise_stats().lrms

            if m == "npkr":
                return noise_stats().pkr

            if m == "ntr":
                return noise_stats().tr

            if m == "ncr":
                return noise_stats().cr

            if m == "nfl":
                return noise_stats().fl

            if m == "npkc":
                return noise_stats().pkc

            if m == "sedavg":
                return 0

            if m == "sedcnt":
                return 0

            if m == "sedtop3":
                return np.zeros(3, dtype=np.float32)

            if m == "sedtopn":
                return 0

            if m == "ssnr":
                return segsnr_f()

            raise AttributeError(f"Unrecognized metric: '{m}'")

        result: dict[str, Any] = {}
        for metric in metrics:
            result[metric] = calc(metric)

            # Check for metrics dependencies and add them even if not explicitly requested.
            if metric.startswith("mxwer"):
                dependencies = ("mxasr." + metric[6:], "tasr." + metric[6:])
                for dependency in dependencies:
                    result[dependency] = calc(dependency)

        return result


def _spectral_mask(db: partial, sm_id: int, use_cache: bool = True) -> SpectralMask:
    """Get spectral mask with ID from db

    :param db: Database context
    :param sm_id: Spectral mask ID
    :param use_cache: If true, use LRU caching
    :return: Spectral mask
    """
    if use_cache:
        return __spectral_mask(db, sm_id)
    return __spectral_mask.__wrapped__(db, sm_id)


@lru_cache
def __spectral_mask(db: partial, sm_id: int) -> SpectralMask:
    from .db_datatypes import SpectralMaskRecord

    with db() as c:
        spectral_mask = SpectralMaskRecord(
            *c.execute(
                """
                SELECT *
                FROM spectral_mask
                WHERE ? = spectral_mask.id
                """,
                (sm_id,),
            ).fetchone()
        )
        return SpectralMask(
            f_max_width=spectral_mask.f_max_width,
            f_num=spectral_mask.f_num,
            t_max_width=spectral_mask.t_max_width,
            t_num=spectral_mask.t_num,
            t_max_percent=spectral_mask.t_max_percent,
        )


def _target_file(db: partial, t_id: int, use_cache: bool = True) -> TargetFile:
    """Get target file with ID from db

    :param db: Database context
    :param t_id: Target file ID
    :param use_cache: If true, use LRU caching
    :return: Target file
    """
    if use_cache:
        return __target_file(db, t_id, use_cache)
    return __target_file.__wrapped__(db, t_id, use_cache)


@lru_cache
def __target_file(db: partial, t_id: int, use_cache: bool = True) -> TargetFile:
    """Get target file with ID from db

    :param db: Database context
    :param t_id: Target file ID
    :param use_cache: If true, use LRU caching
    :return: Target file
    """
    import json

    from .db_datatypes import TargetFileRecord

    with db() as c:
        target_file = TargetFileRecord(
            *c.execute(
                """
                SELECT *
                FROM target_file
                WHERE ? = target_file.id
                """,
                (t_id,),
            ).fetchone()
        )

        return TargetFile(
            name=target_file.name,
            samples=target_file.samples,
            class_indices=json.loads(target_file.class_indices),
            level_type=target_file.level_type,
            truth_configs=_target_truth_configs(db, t_id, use_cache),
            speaker_id=target_file.speaker_id,
        )


def _noise_file(db: partial, n_id: int, use_cache: bool = True) -> NoiseFile:
    """Get noise file with ID from db

    :param db: Database context
    :param n_id: Noise file ID
    :param use_cache: If true, use LRU caching
    :return: Noise file
    """
    if use_cache:
        return __noise_file(db, n_id)
    return __noise_file.__wrapped__(db, n_id)


@lru_cache
def __noise_file(db: partial, n_id: int) -> NoiseFile:
    with db() as c:
        noise = c.execute(
            """
            SELECT noise_file.name, samples
            FROM noise_file
            WHERE ? = noise_file.id
            """,
            (n_id,),
        ).fetchone()
        return NoiseFile(name=noise[0], samples=noise[1])


def _impulse_response_file(db: partial, ir_id: int, use_cache: bool = True) -> str:
    """Get impulse response file with ID from db

    :param db: Database context
    :param ir_id: Impulse response file ID
    :param use_cache: If true, use LRU caching
    :return: Impulse response
    """
    if use_cache:
        return __impulse_response_file(db, ir_id)
    return __impulse_response_file.__wrapped__(db, ir_id)


@lru_cache
def __impulse_response_file(db: partial, ir_id: int) -> str:
    with db() as c:
        return str(
            c.execute(
                """
                SELECT impulse_response_file.file
                FROM impulse_response_file
                WHERE ? = impulse_response_file.id
                """,
                (ir_id + 1,),
            ).fetchone()[0]
        )


def _mixture(db: partial, m_id: int, use_cache: bool = True) -> Mixture:
    """Get mixture record with ID from db

    :param db: Database context
    :param m_id: Zero-based mixture ID
    :param use_cache: If true, use LRU caching
    :return: Mixture record
    """
    if use_cache:
        return __mixture(db, m_id)
    return __mixture.__wrapped__(db, m_id)


@lru_cache
def __mixture(db: partial, m_id: int) -> Mixture:
    from .db_datatypes import MixtureRecord
    from .db_datatypes import TargetRecord
    from .helpers import to_mixture
    from .helpers import to_target

    with db() as c:
        mixture = MixtureRecord(
            *c.execute(
                """
                SELECT *
                FROM mixture
                WHERE ? = mixture.id
                """,
                (m_id + 1,),
            ).fetchone()
        )

        targets = [
            to_target(TargetRecord(*target))
            for target in c.execute(
                """
                SELECT target.*
                FROM target, mixture_target
                WHERE ? = mixture_target.mixture_id AND target.id = mixture_target.target_id
                """,
                (mixture.id,),
            ).fetchall()
        ]

        return to_mixture(mixture, targets)


def _speaker(db: partial, s_id: int | None, tier: str, use_cache: bool = True) -> str | None:
    if use_cache:
        return __speaker(db, s_id, tier)
    return __speaker.__wrapped__(db, s_id, tier)


@lru_cache
def __speaker(db: partial, s_id: int | None, tier: str) -> str | None:
    if s_id is None:
        return None

    with db() as c:
        data = c.execute(f"SELECT {tier} FROM speaker WHERE ? = id", (s_id,)).fetchone()
        if data is None:
            return None
        if data[0] is None:
            return None
        return data[0]


def _target_truth_configs(db: partial, t_id: int, use_cache: bool = True) -> TruthConfigs:
    if use_cache:
        return __target_truth_configs(db, t_id)
    return __target_truth_configs.__wrapped__(db, t_id)


@lru_cache
def __target_truth_configs(db: partial, t_id: int) -> TruthConfigs:
    import json

    from .datatypes import TruthConfig

    truth_configs: TruthConfigs = {}
    with db() as c:
        for truth_config_record in c.execute(
            """
            SELECT truth_config.config
            FROM truth_config, target_file_truth_config
            WHERE ? = target_file_truth_config.target_file_id AND truth_config.id = target_file_truth_config.truth_config_id
            """,
            (t_id,),
        ).fetchall():
            truth_config = json.loads(truth_config_record[0])
            truth_configs[truth_config["name"]] = TruthConfig(
                function=truth_config["function"],
                stride_reduction=truth_config["stride_reduction"],
                config=truth_config["config"],
            )
    return truth_configs



# ruff: noqa: S608
from .datatypes import AudioT
from .datatypes import Augmentation
from .datatypes import AugmentationRule
from .datatypes import AugmentedTarget
from .datatypes import GenMixData
from .datatypes import ImpulseResponseFile
from .datatypes import Mixture
from .datatypes import NoiseFile
from .datatypes import SpectralMask
from .datatypes import Target
from .datatypes import TargetFile
from .datatypes import UniversalSNRGenerator
from .mixdb import MixtureDatabase


def config_file(location: str) -> str:
    from os.path import join

    return join(location, "config.yml")


def initialize_db(location: str, test: bool = False) -> None:
    from .mixdb import db_connection

    con = db_connection(location=location, create=True, test=test)

    con.execute("""
    CREATE TABLE truth_config(
    id INTEGER PRIMARY KEY NOT NULL,
    config TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE truth_parameters(
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    parameters INTEGER)
    """)

    con.execute("""
    CREATE TABLE target_file (
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL,
    class_indices TEXT NOT NULL,
    level_type TEXT NOT NULL,
    speaker_id INTEGER,
    FOREIGN KEY(speaker_id) REFERENCES speaker (id))
    """)

    con.execute("""
    CREATE TABLE speaker (
    id INTEGER PRIMARY KEY NOT NULL,
    parent TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE noise_file (
    id INTEGER PRIMARY KEY NOT NULL,
    name TEXT NOT NULL,
    samples INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE top (
    id INTEGER PRIMARY KEY NOT NULL,
    version INTEGER NOT NULL,
    asr_configs TEXT NOT NULL,
    class_balancing BOOLEAN NOT NULL,
    feature TEXT NOT NULL,
    noise_mix_mode TEXT NOT NULL,
    num_classes INTEGER NOT NULL,
    seed INTEGER NOT NULL,
    mixid_width INTEGER NOT NULL,
    speaker_metadata_tiers TEXT NOT NULL,
    textgrid_metadata_tiers TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_label (
    id INTEGER PRIMARY KEY NOT NULL,
    label TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE class_weights_threshold (
    id INTEGER PRIMARY KEY NOT NULL,
    threshold FLOAT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE impulse_response_file (
    id INTEGER PRIMARY KEY NOT NULL,
    file TEXT NOT NULL,
    tags TEXT NOT NULL)
    """)

    con.execute("""
    CREATE TABLE spectral_mask (
    id INTEGER PRIMARY KEY NOT NULL,
    f_max_width INTEGER NOT NULL,
    f_num INTEGER NOT NULL,
    t_max_width INTEGER NOT NULL,
    t_num INTEGER NOT NULL,
    t_max_percent INTEGER NOT NULL)
    """)

    con.execute("""
    CREATE TABLE target_file_truth_config (
    target_file_id INTEGER,
    truth_config_id INTEGER,
    FOREIGN KEY(target_file_id) REFERENCES target_file (id),
    FOREIGN KEY(truth_config_id) REFERENCES truth_config (id))
    """)

    con.execute("""
    CREATE TABLE target (
    id INTEGER PRIMARY KEY NOT NULL,
    file_id INTEGER NOT NULL,
    augmentation TEXT NOT NULL,
    FOREIGN KEY(file_id) REFERENCES target_file (id))
    """)

    con.execute("""
    CREATE TABLE mixture (
    id INTEGER PRIMARY KEY NOT NULL,
    name VARCHAR NOT NULL,
    noise_file_id INTEGER NOT NULL,
    noise_augmentation TEXT NOT NULL,
    noise_offset INTEGER NOT NULL,
    noise_snr_gain FLOAT,
    random_snr BOOLEAN NOT NULL,
    snr FLOAT NOT NULL,
    samples INTEGER NOT NULL,
    spectral_mask_id INTEGER NOT NULL,
    spectral_mask_seed INTEGER NOT NULL,
    target_snr_gain FLOAT,
    FOREIGN KEY(noise_file_id) REFERENCES noise_file (id),
    FOREIGN KEY(spectral_mask_id) REFERENCES spectral_mask (id))
    """)

    con.execute("""
    CREATE TABLE mixture_target (
    mixture_id INTEGER,
    target_id INTEGER,
    FOREIGN KEY(mixture_id) REFERENCES mixture (id),
    FOREIGN KEY(target_id) REFERENCES target (id))
    """)

    con.commit()
    con.close()


def populate_top_table(location: str, config: dict, test: bool = False) -> None:
    """Populate top table"""
    import json

    from .constants import MIXDB_VERSION
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.execute(
        """
    INSERT INTO top (id, version, asr_configs, class_balancing, feature, noise_mix_mode, num_classes,
    seed, mixid_width, speaker_metadata_tiers, textgrid_metadata_tiers)
    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """,
        (
            1,
            MIXDB_VERSION,
            json.dumps(config["asr_configs"]),
            config["class_balancing"],
            config["feature"],
            config["noise_mix_mode"],
            config["num_classes"],
            config["seed"],
            0,
            "",
            "",
        ),
    )
    con.commit()
    con.close()


def populate_class_label_table(location: str, config: dict, test: bool = False) -> None:
    """Populate class_label table"""
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        "INSERT INTO class_label (label) VALUES (?)",
        [(item,) for item in config["class_labels"]],
    )
    con.commit()
    con.close()


def populate_class_weights_threshold_table(location: str, config: dict, test: bool = False) -> None:
    """Populate class_weights_threshold table"""
    from .mixdb import db_connection

    class_weights_threshold = config["class_weights_threshold"]
    num_classes = config["num_classes"]

    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold]

    if len(class_weights_threshold) == 1:
        class_weights_threshold = [class_weights_threshold[0]] * num_classes

    if len(class_weights_threshold) != num_classes:
        raise ValueError(f"invalid class_weights_threshold length: {len(class_weights_threshold)}")

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        "INSERT INTO class_weights_threshold (threshold) VALUES (?)",
        [(item,) for item in class_weights_threshold],
    )
    con.commit()
    con.close()


def populate_spectral_mask_table(location: str, config: dict, test: bool = False) -> None:
    """Populate spectral_mask table"""
    from .config import get_spectral_masks
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        """
    INSERT INTO spectral_mask (f_max_width, f_num, t_max_width, t_num, t_max_percent) VALUES (?, ?, ?, ?, ?)
    """,
        [
            (
                item.f_max_width,
                item.f_num,
                item.t_max_width,
                item.t_num,
                item.t_max_percent,
            )
            for item in get_spectral_masks(config)
        ],
    )
    con.commit()
    con.close()


def populate_truth_parameters_table(location: str, config: dict, test: bool = False) -> None:
    """Populate truth_parameters table"""
    from .config import get_truth_parameters
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        """
    INSERT INTO truth_parameters (name, parameters) VALUES (?, ?)
    """,
        [
            (
                item.name,
                item.parameters,
            )
            for item in get_truth_parameters(config)
        ],
    )
    con.commit()
    con.close()


def populate_target_file_table(location: str, target_files: list[TargetFile], test: bool = False) -> None:
    """Populate target file table"""
    import json
    from pathlib import Path

    from .mixdb import db_connection

    _populate_truth_config_table(location, target_files, test)
    _populate_speaker_table(location, target_files, test)

    con = db_connection(location=location, readonly=False, test=test)

    cur = con.cursor()
    textgrid_metadata_tiers: set[str] = set()
    for target_file in target_files:
        # Get TextGrid tiers for target file and add to collection
        tiers = _get_textgrid_tiers_from_target_file(target_file.name)
        for tier in tiers:
            textgrid_metadata_tiers.add(tier)

        # Get truth settings for target file
        truth_config_ids: list[int] = []
        for name, config in target_file.truth_configs.items():
            ts = json.dumps({"name": name} | config.to_dict())
            cur.execute(
                "SELECT truth_config.id FROM truth_config WHERE ? = truth_config.config",
                (ts,),
            )
            truth_config_ids.append(cur.fetchone()[0])

        # Get speaker_id for target file
        cur.execute(
            "SELECT speaker.id FROM speaker WHERE ? = speaker.parent",
            (Path(target_file.name).parent.as_posix(),),
        )
        result = cur.fetchone()
        speaker_id = None
        if result is not None:
            speaker_id = result[0]

        # Add entry
        cur.execute(
            "INSERT INTO target_file (name, samples, class_indices, level_type, speaker_id) VALUES (?, ?, ?, ?, ?)",
            (
                target_file.name,
                target_file.samples,
                json.dumps(target_file.class_indices),
                target_file.level_type,
                speaker_id,
            ),
        )
        target_file_id = cur.lastrowid
        for truth_config_id in truth_config_ids:
            cur.execute(
                "INSERT INTO target_file_truth_config (target_file_id, truth_config_id) VALUES (?, ?)",
                (target_file_id, truth_config_id),
            )

    # Update textgrid_metadata_tiers in the top table
    con.execute(
        "UPDATE top SET textgrid_metadata_tiers=? WHERE ? = top.id",
        (json.dumps(sorted(textgrid_metadata_tiers)), 1),
    )

    con.commit()
    con.close()


def populate_noise_file_table(location: str, noise_files: list[NoiseFile], test: bool = False) -> None:
    """Populate noise file table"""
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        "INSERT INTO noise_file (name, samples) VALUES (?, ?)",
        [(noise_file.name, noise_file.samples) for noise_file in noise_files],
    )
    con.commit()
    con.close()


def populate_impulse_response_file_table(
    location: str, impulse_response_files: list[ImpulseResponseFile], test: bool = False
) -> None:
    """Populate impulse response file table"""
    import json

    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.executemany(
        "INSERT INTO impulse_response_file (file, tags) VALUES (?, ?)",
        [
            (
                impulse_response_file.file,
                json.dumps(impulse_response_file.tags),
            )
            for impulse_response_file in impulse_response_files
        ],
    )
    con.commit()
    con.close()


def update_mixid_width(location: str, num_mixtures: int, test: bool = False) -> None:
    """Update the mixid width"""
    from sonusai.utils import max_text_width

    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)
    con.execute(
        "UPDATE top SET mixid_width=? WHERE ? = top.id",
        (max_text_width(num_mixtures), 1),
    )
    con.commit()
    con.close()


def generate_mixtures(
    noise_mix_mode: str,
    augmented_targets: list[AugmentedTarget],
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    spectral_masks: list[SpectralMask],
    all_snrs: list[UniversalSNRGenerator],
    mixups: list[int],
    num_classes: int,
    feature_step_samples: int,
    num_ir: int,
) -> tuple[int, int, list[Mixture]]:
    """Generate mixtures"""
    if noise_mix_mode == "exhaustive":
        func = _exhaustive_noise_mix
    elif noise_mix_mode == "non-exhaustive":
        func = _non_exhaustive_noise_mix
    elif noise_mix_mode == "non-combinatorial":
        func = _non_combinatorial_noise_mix
    else:
        raise ValueError(f"invalid noise_mix_mode: {noise_mix_mode}")

    return func(
        augmented_targets=augmented_targets,
        target_files=target_files,
        target_augmentations=target_augmentations,
        noise_files=noise_files,
        noise_augmentations=noise_augmentations,
        spectral_masks=spectral_masks,
        all_snrs=all_snrs,
        mixups=mixups,
        num_classes=num_classes,
        feature_step_samples=feature_step_samples,
        num_ir=num_ir,
    )


def populate_mixture_table(
    location: str,
    mixtures: list[Mixture],
    test: bool = False,
    logging: bool = False,
    show_progress: bool = False,
) -> None:
    """Populate mixture table"""
    from sonusai import logger
    from sonusai.utils import track

    from .helpers import from_mixture
    from .helpers import from_target
    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)

    # Populate target table
    if logging:
        logger.info("Populating target table")
    targets: list[tuple[int, str]] = []
    for mixture in mixtures:
        for target in mixture.targets:
            entry = from_target(target)
            if entry not in targets:
                targets.append(entry)
    for target in track(targets, disable=not show_progress):
        con.execute("INSERT INTO target (file_id, augmentation) VALUES (?, ?)", target)

    # Populate mixture table
    if logging:
        logger.info("Populating mixture table")
    for mixture in track(mixtures, disable=not show_progress):
        m_id = int(mixture.name)
        con.execute(
            """
            INSERT INTO mixture (id, name, noise_file_id, noise_augmentation, noise_offset, noise_snr_gain, random_snr,
            snr, samples, spectral_mask_id, spectral_mask_seed, target_snr_gain)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
            (m_id + 1, *from_mixture(mixture)),
        )

        for target in mixture.targets:
            target_id = con.execute(
                """
                SELECT target.id
                FROM target
                WHERE ? = target.file_id AND ? = target.augmentation
            """,
                from_target(target),
            ).fetchone()[0]
            con.execute(
                "INSERT INTO mixture_target (mixture_id, target_id) VALUES (?, ?)",
                (m_id + 1, target_id),
            )

    con.commit()
    con.close()


def update_mixture(mixdb: MixtureDatabase, mixture: Mixture, with_data: bool = False) -> tuple[Mixture, GenMixData]:
    """Update mixture record with name and gains"""
    from .audio import get_next_noise
    from .augmentation import apply_gain
    from .helpers import get_target

    mixture, targets_audio = _initialize_targets_audio(mixdb, mixture)

    noise_audio = _augmented_noise_audio(mixdb, mixture)
    noise_audio = get_next_noise(audio=noise_audio, offset=mixture.noise_offset, length=mixture.samples)

    # Apply IR and sum targets audio before initializing the mixture SNR gains
    target_audio = get_target(mixdb, mixture, targets_audio)

    mixture = _initialize_mixture_gains(
        mixdb=mixdb, mixture=mixture, target_audio=target_audio, noise_audio=noise_audio
    )

    mixture.name = f"{int(mixture.name):0{mixdb.mixid_width}}"

    if not with_data:
        return mixture, GenMixData()

    # Apply SNR gains
    targets_audio = [apply_gain(audio=target_audio, gain=mixture.target_snr_gain) for target_audio in targets_audio]
    noise_audio = apply_gain(audio=noise_audio, gain=mixture.noise_snr_gain)

    # Apply IR and sum targets audio after applying the mixture SNR gains
    target_audio = get_target(mixdb, mixture, targets_audio)
    mixture_audio = target_audio + noise_audio

    return mixture, GenMixData(
        mixture=mixture_audio,
        targets=targets_audio,
        target=target_audio,
        noise=noise_audio,
    )


def _augmented_noise_audio(mixdb: MixtureDatabase, mixture: Mixture) -> AudioT:
    from .audio import read_audio
    from .audio import read_ir
    from .augmentation import apply_augmentation
    from .augmentation import apply_impulse_response

    noise = mixdb.noise_file(mixture.noise.file_id)
    noise_augmentation = mixture.noise.augmentation

    audio = read_audio(noise.name)
    audio = apply_augmentation(audio, noise_augmentation)
    if noise_augmentation.ir is not None:
        audio = apply_impulse_response(audio, read_ir(mixdb.impulse_response_file(noise_augmentation.ir)))  # pyright: ignore [reportArgumentType]

    return audio


def _initialize_targets_audio(mixdb: MixtureDatabase, mixture: Mixture) -> tuple[Mixture, list[AudioT]]:
    from .augmentation import apply_augmentation
    from .augmentation import pad_audio_to_length

    targets_audio = []
    for target in mixture.targets:
        target_audio = mixdb.read_target_audio(target.file_id)
        targets_audio.append(
            apply_augmentation(
                audio=target_audio,
                augmentation=target.augmentation,
                frame_length=mixdb.feature_step_samples,
            )
        )

    mixture.samples = max([len(item) for item in targets_audio])

    for idx in range(len(targets_audio)):
        targets_audio[idx] = pad_audio_to_length(audio=targets_audio[idx], length=mixture.samples)

    return mixture, targets_audio


def _initialize_mixture_gains(
    mixdb: MixtureDatabase,
    mixture: Mixture,
    target_audio: AudioT,
    noise_audio: AudioT,
) -> Mixture:
    import numpy as np

    from sonusai.utils import asl_p56
    from sonusai.utils import db_to_linear

    if mixture.is_noise_only:
        # Special case for zeroing out target data
        mixture.target_snr_gain = 0
        mixture.noise_snr_gain = 1
    elif mixture.is_target_only:
        # Special case for zeroing out noise data
        mixture.target_snr_gain = 1
        mixture.noise_snr_gain = 0
    else:
        target_level_types = [
            target_file.level_type for target_file in [mixdb.target_file(target.file_id) for target in mixture.targets]
        ]
        if not all(level_type == target_level_types[0] for level_type in target_level_types):
            raise ValueError("Not all target_level_types in mixup are the same")

        level_type = target_level_types[0]
        match level_type:
            case "default":
                target_energy = np.mean(np.square(target_audio))
            case "speech":
                target_energy = asl_p56(target_audio)
            case _:
                raise ValueError(f"Unknown level_type: {level_type}")

        noise_energy = np.mean(np.square(noise_audio))
        if noise_energy == 0:
            noise_gain = 1
        else:
            noise_gain = np.sqrt(target_energy / noise_energy) / db_to_linear(mixture.snr)

        # Check for noise_gain > 1 to avoid clipping
        if noise_gain > 1:
            mixture.target_snr_gain = 1 / noise_gain
            mixture.noise_snr_gain = 1
        else:
            mixture.target_snr_gain = 1
            mixture.noise_snr_gain = noise_gain

    # Check for clipping in mixture
    gain_adjusted_target_audio = target_audio * mixture.target_snr_gain
    gain_adjusted_noise_audio = noise_audio * mixture.noise_snr_gain
    mixture_audio = gain_adjusted_target_audio + gain_adjusted_noise_audio
    max_abs_audio = max(abs(mixture_audio))
    clip_level = db_to_linear(-0.25)
    if max_abs_audio > clip_level:
        # Clipping occurred; lower gains to bring audio within +/-1
        gain_adjustment = clip_level / max_abs_audio
        mixture.target_snr_gain *= gain_adjustment
        mixture.noise_snr_gain *= gain_adjustment

    mixture.target_snr_gain = round(mixture.target_snr_gain, ndigits=5)
    mixture.noise_snr_gain = round(mixture.noise_snr_gain, ndigits=5)
    return mixture


def _exhaustive_noise_mix(
    augmented_targets: list[AugmentedTarget],
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    spectral_masks: list[SpectralMask],
    all_snrs: list[UniversalSNRGenerator],
    mixups: list[int],
    num_classes: int,
    feature_step_samples: int,
    num_ir: int,
) -> tuple[int, int, list[Mixture]]:
    """Use every noise/augmentation with every target/augmentation+interferences/augmentation"""
    from random import randint

    import numpy as np

    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length
    from .datatypes import Mixture
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    m_id = 0
    used_noise_files = len(noise_files) * len(noise_augmentations)
    used_noise_samples = 0

    augmented_target_ids_for_mixups = [
        get_augmented_target_ids_for_mixup(
            augmented_targets=augmented_targets,
            targets=target_files,
            target_augmentations=target_augmentations,
            mixup=mixup,
            num_classes=num_classes,
        )
        for mixup in mixups
    ]

    mixtures: list[Mixture] = []
    for noise_file_id in range(len(noise_files)):
        for noise_augmentation_rule in noise_augmentations:
            noise_augmentation = augmentation_from_rule(noise_augmentation_rule, num_ir)
            noise_offset = 0
            noise_length = estimate_augmented_length_from_length(
                length=noise_files[noise_file_id].samples,
                tempo=noise_augmentation.tempo,
            )

            for augmented_target_ids_for_mixup in augmented_target_ids_for_mixups:
                for augmented_target_ids in augmented_target_ids_for_mixup:
                    targets, target_length = _get_target_info(
                        augmented_target_ids=augmented_target_ids,
                        augmented_targets=augmented_targets,
                        target_files=target_files,
                        target_augmentations=target_augmentations,
                        feature_step_samples=feature_step_samples,
                        num_ir=num_ir,
                    )

                    for spectral_mask_id in range(len(spectral_masks)):
                        for snr in all_snrs:
                            mixtures.append(
                                Mixture(
                                    targets=targets,
                                    name=str(m_id),
                                    noise=Noise(file_id=noise_file_id + 1, augmentation=noise_augmentation),
                                    noise_offset=noise_offset,
                                    samples=target_length,
                                    snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                                    spectral_mask_id=spectral_mask_id + 1,
                                    spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                                )
                            )
                            m_id += 1

                            noise_offset = int((noise_offset + target_length) % noise_length)
                            used_noise_samples += target_length

    return used_noise_files, used_noise_samples, mixtures


def _non_exhaustive_noise_mix(
    augmented_targets: list[AugmentedTarget],
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    spectral_masks: list[SpectralMask],
    all_snrs: list[UniversalSNRGenerator],
    mixups: list[int],
    num_classes: int,
    feature_step_samples: int,
    num_ir: int,
) -> tuple[int, int, list[Mixture]]:
    """Cycle through every target/augmentation+interferences/augmentation without necessarily using all
    noise/augmentation combinations (reduced data set).
    """
    from random import randint

    import numpy as np

    from .datatypes import Mixture
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    m_id = 0
    used_noise_files = set()
    used_noise_samples = 0
    noise_file_id = None
    noise_augmentation_id = None
    noise_offset = None

    augmented_target_indices_for_mixups = [
        get_augmented_target_ids_for_mixup(
            augmented_targets=augmented_targets,
            targets=target_files,
            target_augmentations=target_augmentations,
            mixup=mixup,
            num_classes=num_classes,
        )
        for mixup in mixups
    ]

    mixtures: list[Mixture] = []
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            targets, target_length = _get_target_info(
                augmented_target_ids=augmented_target_indices,
                augmented_targets=augmented_targets,
                target_files=target_files,
                target_augmentations=target_augmentations,
                feature_step_samples=feature_step_samples,
                num_ir=num_ir,
            )

            for spectral_mask_id in range(len(spectral_masks)):
                for snr in all_snrs:
                    (
                        noise_file_id,
                        noise_augmentation_id,
                        noise_augmentation,
                        noise_offset,
                    ) = _get_next_noise_offset(
                        noise_file_id=noise_file_id,
                        noise_augmentation_id=noise_augmentation_id,
                        noise_offset=noise_offset,
                        target_length=target_length,
                        noise_files=noise_files,
                        noise_augmentations=noise_augmentations,
                        num_ir=num_ir,
                    )
                    used_noise_samples += target_length

                    used_noise_files.add(f"{noise_file_id}_{noise_augmentation_id}")

                    mixtures.append(
                        Mixture(
                            targets=targets,
                            name=str(m_id),
                            noise=Noise(file_id=noise_file_id + 1, augmentation=noise_augmentation),
                            noise_offset=noise_offset,
                            samples=target_length,
                            snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                            spectral_mask_id=spectral_mask_id + 1,
                            spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                        )
                    )
                    m_id += 1

    return len(used_noise_files), used_noise_samples, mixtures


def _non_combinatorial_noise_mix(
    augmented_targets: list[AugmentedTarget],
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    spectral_masks: list[SpectralMask],
    all_snrs: list[UniversalSNRGenerator],
    mixups: list[int],
    num_classes: int,
    feature_step_samples: int,
    num_ir: int,
) -> tuple[int, int, list[Mixture]]:
    """Combine a target/augmentation+interferences/augmentation with a single cut of a noise/augmentation
    non-exhaustively (each target/augmentation+interferences/augmentation does not use each noise/augmentation).
    Cut has random start and loop back to beginning if end of noise/augmentation is reached.
    """
    from random import choice
    from random import randint

    import numpy as np

    from .datatypes import Mixture
    from .datatypes import Noise
    from .datatypes import UniversalSNR
    from .targets import get_augmented_target_ids_for_mixup

    m_id = 0
    used_noise_files = set()
    used_noise_samples = 0
    noise_file_id = None
    noise_augmentation_id = None

    augmented_target_indices_for_mixups = [
        get_augmented_target_ids_for_mixup(
            augmented_targets=augmented_targets,
            targets=target_files,
            target_augmentations=target_augmentations,
            mixup=mixup,
            num_classes=num_classes,
        )
        for mixup in mixups
    ]

    mixtures: list[Mixture] = []
    for mixup in augmented_target_indices_for_mixups:
        for augmented_target_indices in mixup:
            targets, target_length = _get_target_info(
                augmented_target_ids=augmented_target_indices,
                augmented_targets=augmented_targets,
                target_files=target_files,
                target_augmentations=target_augmentations,
                feature_step_samples=feature_step_samples,
                num_ir=num_ir,
            )

            for spectral_mask_id in range(len(spectral_masks)):
                for snr in all_snrs:
                    (
                        noise_file_id,
                        noise_augmentation_id,
                        noise_augmentation,
                        noise_length,
                    ) = _get_next_noise_indices(
                        noise_file_id=noise_file_id,
                        noise_augmentation_id=noise_augmentation_id,
                        noise_files=noise_files,
                        noise_augmentations=noise_augmentations,
                        num_ir=num_ir,
                    )
                    used_noise_samples += target_length

                    used_noise_files.add(f"{noise_file_id}_{noise_augmentation_id}")

                    mixtures.append(
                        Mixture(
                            targets=targets,
                            name=str(m_id),
                            noise=Noise(file_id=noise_file_id + 1, augmentation=noise_augmentation),
                            noise_offset=choice(range(noise_length)),  # noqa: S311
                            samples=target_length,
                            snr=UniversalSNR(value=snr.value, is_random=snr.is_random),
                            spectral_mask_id=spectral_mask_id + 1,
                            spectral_mask_seed=randint(0, np.iinfo("i").max),  # noqa: S311
                        )
                    )
                    m_id += 1

    return len(used_noise_files), used_noise_samples, mixtures


def _get_next_noise_indices(
    noise_file_id: int | None,
    noise_augmentation_id: int | None,
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    num_ir: int,
) -> tuple[int, int, Augmentation, int]:
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length

    if noise_file_id is None or noise_augmentation_id is None:
        noise_file_id = 0
        noise_augmentation_id = 0
    else:
        noise_augmentation_id += 1
        if noise_augmentation_id == len(noise_augmentations):
            noise_augmentation_id = 0
            noise_file_id += 1
            if noise_file_id == len(noise_files):
                noise_file_id = 0

    noise_augmentation = augmentation_from_rule(noise_augmentations[noise_augmentation_id], num_ir)
    noise_length = estimate_augmented_length_from_length(
        length=noise_files[noise_file_id].samples, tempo=noise_augmentation.tempo
    )
    return noise_file_id, noise_augmentation_id, noise_augmentation, noise_length


def _get_next_noise_offset(
    noise_file_id: int | None,
    noise_augmentation_id: int | None,
    noise_offset: int | None,
    target_length: int,
    noise_files: list[NoiseFile],
    noise_augmentations: list[AugmentationRule],
    num_ir: int,
) -> tuple[int, int, Augmentation, int]:
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length

    if noise_file_id is None or noise_augmentation_id is None or noise_offset is None:
        noise_file_id = 0
        noise_augmentation_id = 0
        noise_offset = 0

    noise_augmentation = augmentation_from_rule(noise_augmentations[noise_file_id], num_ir)
    noise_length = estimate_augmented_length_from_length(
        length=noise_files[noise_file_id].samples, tempo=noise_augmentation.tempo
    )
    if noise_offset + target_length >= noise_length:
        if noise_offset == 0:
            raise ValueError("Length of target audio exceeds length of noise audio")

        noise_offset = 0
        noise_augmentation_id += 1
        if noise_augmentation_id == len(noise_augmentations):
            noise_augmentation_id = 0
            noise_file_id += 1
            if noise_file_id == len(noise_files):
                noise_file_id = 0
        noise_augmentation = augmentation_from_rule(noise_augmentations[noise_augmentation_id], num_ir)

    return noise_file_id, noise_augmentation_id, noise_augmentation, noise_offset


def _get_target_info(
    augmented_target_ids: list[int],
    augmented_targets: list[AugmentedTarget],
    target_files: list[TargetFile],
    target_augmentations: list[AugmentationRule],
    feature_step_samples: int,
    num_ir: int,
) -> tuple[list[Target], int]:
    from .augmentation import augmentation_from_rule
    from .augmentation import estimate_augmented_length_from_length

    mixups: list[Target] = []
    target_length = 0
    for idx in augmented_target_ids:
        tfi = augmented_targets[idx].target_id
        target_augmentation_rule = target_augmentations[augmented_targets[idx].target_augmentation_id]
        target_augmentation = augmentation_from_rule(target_augmentation_rule, num_ir)

        mixups.append(Target(file_id=tfi + 1, augmentation=target_augmentation))

        target_length = max(
            estimate_augmented_length_from_length(
                length=target_files[tfi].samples,
                tempo=target_augmentation.tempo,
                frame_length=feature_step_samples,
            ),
            target_length,
        )
    return mixups, target_length


def get_all_snrs_from_config(config: dict) -> list[UniversalSNRGenerator]:
    from .datatypes import UniversalSNRGenerator

    return [UniversalSNRGenerator(is_random=False, _raw_value=snr) for snr in config["snrs"]] + [
        UniversalSNRGenerator(is_random=True, _raw_value=snr) for snr in config["random_snrs"]
    ]


def _get_textgrid_tiers_from_target_file(target_file: str) -> list[str]:
    from pathlib import Path

    from praatio import textgrid

    from sonusai.mixture import tokenized_expand

    textgrid_file = Path(tokenized_expand(target_file)[0]).with_suffix(".TextGrid")
    if not textgrid_file.exists():
        return []

    tg = textgrid.openTextgrid(str(textgrid_file), includeEmptyIntervals=False)

    return sorted(tg.tierNames)


def _populate_speaker_table(location: str, target_files: list[TargetFile], test: bool = False) -> None:
    """Populate speaker table"""
    import json
    from pathlib import Path

    import yaml

    from .mixdb import db_connection
    from .tokenized_shell_vars import tokenized_expand

    # Determine columns for speaker table
    all_parents = {Path(target_file.name).parent for target_file in target_files}
    speaker_parents = (parent for parent in all_parents if Path(tokenized_expand(parent / "speaker.yml")[0]).exists())

    speakers: dict[Path, dict[str, str]] = {}
    for parent in sorted(speaker_parents):
        with open(tokenized_expand(parent / "speaker.yml")[0]) as f:
            speakers[parent] = yaml.safe_load(f)

    new_columns: list[str] = []
    for keys in speakers:
        for column in speakers[keys]:
            new_columns.append(column)
    new_columns = sorted(set(new_columns))

    con = db_connection(location=location, readonly=False, test=test)

    for new_column in new_columns:
        con.execute(f"ALTER TABLE speaker ADD COLUMN {new_column} TEXT")

    # Populate speaker table
    speaker_rows: list[tuple[str, ...]] = []
    for key in speakers:
        entry = (speakers[key].get(column, None) for column in new_columns)
        speaker_rows.append((key.as_posix(), *entry))  # type: ignore[arg-type]

    column_ids = ", ".join(["parent", *new_columns])
    column_values = ", ".join(["?"] * (len(new_columns) + 1))
    con.executemany(f"INSERT INTO speaker ({column_ids}) VALUES ({column_values})", speaker_rows)

    con.execute("CREATE INDEX speaker_parent_idx ON speaker (parent)")

    # Update speaker_metadata_tiers in the top table
    tiers = [
        description[0]
        for description in con.execute("SELECT * FROM speaker").description
        if description[0] not in ("id", "parent")
    ]
    con.execute(
        "UPDATE top SET speaker_metadata_tiers=? WHERE ? = top.id",
        (json.dumps(tiers), 1),
    )

    if "speaker_id" in tiers:
        con.execute("CREATE INDEX speaker_speaker_id_idx ON speaker (speaker_id)")

    con.commit()
    con.close()


def _populate_truth_config_table(location: str, target_files: list[TargetFile], test: bool = False) -> None:
    """Populate truth_config table"""
    import json

    from .mixdb import db_connection

    con = db_connection(location=location, readonly=False, test=test)

    # Populate truth_config table
    truth_configs: list[str] = []
    for target_file in target_files:
        for name, config in target_file.truth_configs.items():
            ts = json.dumps({"name": name} | config.to_dict())
            if ts not in truth_configs:
                truth_configs.append(ts)
    con.executemany(
        "INSERT INTO truth_config (config) VALUES (?)",
        [(item,) for item in truth_configs],
    )

    con.commit()
    con.close()
